// *************************************************************************
//    Compile.C
// *************************************************************************

#include "Compile.h"

#include "CompileRule.h"
#include "Unparse.h"
#include "ParserStruct.h"
#include "Execute.h"
#include "Globals.h"
#include "Literal.h"
#include "SccInfo.h"
#include "BuiltinSymbols.h"
#include "ExportInfo.h"
#include "DerivedMethod.h"
#include "QFPipelinedModuleData.h"
#include "Pipelined.h"
#include "Database.h"

/*****************************************************************************/
int corget_offset(corTable *F, Name name, int arity, int print_error)
{
    for (int i = 0; i< F->numRels; i++) {
        if ((F->rels[i].relArity == arity) && (F->rels[i].relName == name))
            break;
    }
    
    // The offset returned should be within the array bounds.
    if (i >= F->numRels) {
        if (print_error) {
            fprintf(corexEnv->error_file,
                "Error: Couldn't find relation %s with arity %d\n",
                name->string(), arity);
        }
        return(-1);
    }
    ASSERT((i < F->numRels));
    return(i);
}

/*****************************************************************************/
//    NOTE: Predicate field in Literal is often not set !! So use pred field.

int corget_offset(corTable *F, corLiteral *pred, int print_error)
{
    return corget_offset(F, pred->pred, pred->arity(), print_error);
}

/*****************************************************************************/
/*
    This function finds the offset of an RMark in the RMarkArray for an scc
    It depends on the order in which the RMarks are actually allocated, which
    happens in SccInfo::Solve(). The function returns the offset of the first
    RMark for that predicate (i.e.) the old RMark. The new RMark is in the
    adjacent position.
*/

int corFind_RMark_Offset(corSccInfo *scc_info, corPredicate *rpred)
{
    corPredicate **pred_ptrs = scc_info->pred_ptrs;
    int total_numpreds = scc_info->numPreds + scc_info->num_lowerPreds;
    for (int i=0; i < total_numpreds; i++) {
        if (pred_ptrs[i] == rpred) return (2*i);
    }

    // Default --- this should never be reached !!
    fprintf(corexEnv->error_file,
        "CORAL:: Error - Find_RMark_Offset - can't find %s(arity %d)\n",
        rpred->name->string(), rpred->arity);
    return(-1);
}

/*****************************************************************************/
int corhas_fail_in_body(corClause *rule)
{
    int i;

    for (i=0; i < rule->num_preds; i++) {
        if (rule->preds[i]->pred == corFailSymbol)
            return i;
    }
    return -1;
}

/*****************************************************************************/
void coradd_pred_extern(corPredicate *pred, enum PredKind kind,
    corModulePred *predicates)
{
    Name name = pred->name;
    int arity = pred->arity;
    corModulePred *new_pred;

    for (new_pred = predicates->next; new_pred; new_pred = new_pred->next) {
        if (new_pred->pred->name->equals(name) &&
            (new_pred->pred->arity == arity))
        {
            if (kind==LocalPredKind && new_pred->kind==ExternalPredKind)
                new_pred->kind = kind;
            return;
        }
    }

    new_pred = new corModulePred(pred);
    new_pred->kind = kind;
    new_pred->next = predicates->next;
    predicates->next = new_pred;
}

/*****************************************************************************/

corScc *corSccList = NULL;
corScc **corSccListTail = &corSccList;

/*****************************************************************************/

struct corSccWork *corSccWorkList = NULL;
int corSccCount = 0;
struct corSccWork corSccDone[1]; // dummy

/*****************************************************************************/
static struct corScc *NewScc()
{
    struct corScc *scc = new corScc;
    scc->preds = NULL;
    scc->next = NULL;
    scc->recursive = 0;
    *corSccListTail = scc;
    corSccListTail = &scc->next;
    scc->emitted = -1;
    return scc;
}

/*****************************************************************************/
void corinitialize_scc_structs()
{
    corSccList = NULL;
    corSccListTail = &corSccList;
    corSccWorkList = NULL;
    corSccCount = 0;
}

/*****************************************************************************/
inline int min(int i, int j)
{
    return i < j ? i : j;
}

/*****************************************************************************/
int corfind_predicates_invoked_from(corPredicate *pred, int original)
{
    // Based on code by Thomas Ball from an algorithm by Robert Tarjan

    struct corSccWork *work = new corSccWork();
    corClause *clause;
    int ipred;
    // number is a sequential node number
    // lowlink is lowest number of a node reachable from current node.
    work->lowlink = work->number = ++corSccCount;

    work->next = corSccWorkList;
    work->pred = pred;
    work->recursive = 0;
    pred->work = work;
    corSccWorkList = work;
    // Iterate over all Clauses that define 'pred'.
    for (clause = pred->clauses; clause; clause = clause->pnext) {
        // This should take care of 'bogus' magic dependencies
        if (clause->head->predicate->coris_magic() && clause->num_preds) {
            corPredicate *first_pred = clause->preds[0]->predicate;
            if (first_pred->coris_magic() && first_pred->mark == 0)
                continue;
        }

      // when we are constructing the original Scc structure,
      // we ignore rewritten rules, i.e. rules with 2 labels
      // added by Ken
      if (original && (clause->clause_names->next != NULL))
        continue;

        clause->seen = 1;
        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            corLiteral *child_literal = clause->preds[ipred];
            corPredicate *child = child_literal->predicate;
            if (child->work == NULL) { /*  not visited yet */
                int child_lowlink = corfind_predicates_invoked_from(child);
                work->lowlink = min(work->lowlink, child_lowlink);
            }
            else if (child->work != corSccDone) {
                if (child->work->number < work->number)
                    work->lowlink = min(work->lowlink, child->work->lowlink);
                child->work->recursive = 1;
            }
        }
    }
    int lowlink = work->lowlink;
    if (work->lowlink == work->number) {
        struct corScc *scc = NewScc();
        corPredicate **pred_tail = &scc->preds;
        while (corSccWorkList && corSccWorkList->number >= work->number) {
            register corSccWork *sccWork = corSccWorkList;
            corPredicate *work_pred = sccWork->pred;
            work_pred->next = scc->preds;
            work_pred->scc = scc;
#            ifdef DEBUG
                if (work_pred->work != sccWork) abort();
#            endif
            work_pred->work = corSccDone;
            *pred_tail = work_pred;
            pred_tail = &work_pred->next;
            corSccWorkList = sccWork->next;
            scc->recursive |= work->recursive;
            //    delete sccWork;  PURIFY :: this frees memory used
            //    in later recursive calls !! 
        }
        *pred_tail = NULL;
    }
    return lowlink;
}

/*****************************************************************************/
static void mark_predicates_invoked_from(corPredicate *pred)
{
    corClause *clause;
    int ipred;

    pred->mark = 1;
    // Iterate over all Clauses that define 'pred'.
    for (clause = pred->clauses; clause; clause = clause->pnext) {
        if (clause->head->predicate->coris_magic()) continue;

        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            corLiteral *child_literal = clause->preds[ipred];
            corPredicate *child = child_literal->predicate;
            if (child->mark == 0) { // not visited yet
                mark_predicates_invoked_from(child);
            }
        }
    }
}

/*****************************************************************************/
static void clean_up_after_one_scc()
{
    // if there are no base predicates, need to clean up SccList
    struct corScc *base_scc = corSccList;
    if (!(base_scc->preds)) {
        // there are no base predicates
        ASSERT((base_scc->next != NULL));

        // must at least have the query predicate
        ASSERT((base_scc->next->preds != NULL));

        corSccList = base_scc->next;
        delete base_scc;
    }

    // NOTE: should the SccWorkList also be deleted?
}

/*****************************************************************************/
int corgroup_predicates_in_one_scc(corPredicate *pred)
{
    // If there are base predicates, there have to be at least two
    // SCCs; the first one must be the base SCC

    struct corScc *base_scc;
    struct corScc *scc;
    if (!corSccList) {
        base_scc = NewScc();
        scc = NewScc();         // this should link it to be the 2nd SCC
        scc->recursive = 1;        // NOTE: is this a right assumption?
    }
    else {
        base_scc = corSccList;
        scc = corSccList->next;
    }

    struct corSccWork *work;
    if (!corSccWorkList) {
        work = new corSccWork();
        work->next = corSccWorkList;
        corSccWorkList = work;
    }
    else work = corSccWorkList;

    pred->work = work;  // visited

    corClause *clause;
    int ipred;
    
    pred->scc = scc;
    pred->next = scc->preds;
    scc->preds = pred;

    for (clause = pred->clauses; clause; clause = clause->pnext) {
        // This should take care of 'bogus' magic dependencies
        if (clause->head->predicate->coris_magic() && clause->num_preds) {
            corPredicate *first_pred = clause->preds[0]->predicate;
            if (first_pred->coris_magic() && first_pred->mark == 0)
                continue;
        }

        clause->seen = 1;

        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            corLiteral *child_literal = clause->preds[ipred];
            corPredicate *child = child_literal->predicate;
            if ((child->work == NULL) && (child->clauses)) {
                // predicates not yet visited and those
                // which have rules defined in this module
                corgroup_predicates_in_one_scc(child);
            }
            else if (child->work == NULL) {
                // predicates not yet visited
                // and which are external predicates
                child->work = work;
                if (child->coris_mp_done()) { 
                    // NOTE: for Ordered Search
                    child->scc = scc;
                    child->next = scc->preds;
                    scc->preds = child;
                }
                else {
                    child->scc = base_scc;
                    child->next = base_scc->preds;
                    base_scc->preds = child;
                }
            }
        }
    }

    return 1;
}

/*****************************************************************************/
static void interpret_scc(corQFMaterializedModuleData *md, corScc *scc,
    corExportInfo *export_info, corPredicate *ext_magicpred)
{
    corPredicate *pred; corClause *clause;
    int numRules = 0;

    // First emit code to solve Scc's needed by this Scc.
    // (This is, after all, bottom-up evaluation.)

    for (pred = scc->preds; pred != NULL; pred = pred->next) {
        for (clause = pred->clauses;
            clause != *pred->clause_tail;
            clause = clause->pnext) {
            if (!clause->seen) continue;
            int ipred;
            numRules++;
            for (ipred = 1; ipred <= clause->num_preds; ipred++) {
                corPredicate *child = clause->preds[ipred-1]->predicate;
                if (child->scc == scc) continue;
                if (child->scc->emitted == export_info->sequence)
                    continue; // Already dumped.
                child->scc->emitted = export_info->sequence;
                interpret_scc(md, child->scc, export_info, ext_magicpred);
            }
        }
    }

    // Check that the scc has some rules to evaluate. Otherwise, it is useless
    if (numRules)  md->AddScc(scc, ext_magicpred);
}

/*****************************************************************************/
corPredicate *corfind_ext_magicquerypred(corParserStruct& corparserStruct)
{
    for (corPredicate *pred = corparserStruct.AllPredicates; pred != NULL;
        pred=pred->allprednext)
    {
        if ((!pred->clauses) && (pred->coris_magic())) { // found it
            return pred;
        }
    }
    return NULL;
}

/*****************************************************************************/
int corcount_var_list(corVarLink *rest)
{
    corVarLink *cur;
    int max_var_num = -1;
    for (cur=rest;  cur; cur=cur->next) {
        if (cur->arg->var > max_var_num)
            max_var_num = cur->arg->var;
    }
    return max_var_num+1;
}

/*****************************************************************************/
int cordetermine_subsum_status(corQFModuleData *md, Name name, int default_val)
{
    corPredAnnotations *cur;
    
    for (cur= md->multiset_annotations; cur; cur=cur->next()) {
        char buf[1000];
        corsprint_pred_name(buf, corSymbolString(cur->pred), cur->adorn);
        //    Warning: multiset and checksubsumption annotations have
        //    the name stored separate from the adornment 
        if (name->equals(corEnterSymbol(buf))) {
            default_val = cur->check_subsum;
        }
    } 
    /***
    For pipelined modules, this used to be the code : Praveen

    for (cur= md->multiset_annotations; cur; cur=cur->next()) {
        if (cur->pred==rule->head->pred)
             F->rels[offset].relPtr->check_subsum = cur->check_subsum;
    }
    ****/

    return default_val;
}

/*****************************************************************************/
static int diskrel_anno(corQFModuleData *md, Name name, corArgList *& a_list)
{
    corPredAnnotations *cur;

    // what about the schema arglist in this case ?? : PRAVEEN
    if (md->module_info.DiskRel) return 1;
    
    for (cur= md->diskrel_annotations; cur; cur=cur->next()) {
        char buf[1000];
        corsprint_pred_name(buf, corSymbolString(cur->pred), cur->adorn);
        //    Warning: diskrel annotations have the name stored separate 
        //    from the adornment 
        if (corEnterSymbol(buf) == name) {
            a_list = cur->arglist1;
            return 1;
        }
    }
    return 0;
}

/*****************************************************************************/
void corget_annos(corQFModuleData *md, corModulePred *predicate)
{
    corPredAnnotations *cur;
    struct corAnnoStruct *new_anno;

    for (cur= md->index_annotations; cur; cur = cur->next()) {
        if (cur->pred->equals(predicate->pred->name)) {
            corVarLink *rest;
            rest = cur->arglist1->var_list(NULL);
            if (cur->arglist2 )
                rest = cur->arglist2->var_list(rest);
              
            int num_vars = corcount_var_list(rest);
            new_anno = new struct corAnnoStruct(cur, num_vars,
                predicate->index_annos);
            predicate->index_annos = new_anno;
        }
    }

    for (cur= md->agg_sel_annotations; cur; cur=cur->next()) {
        if (cur->pred->equals(predicate->pred->name)) {
            new_anno = new struct corAnnoStruct(cur, 0, predicate->aggsel_annos);
            predicate->aggsel_annos = new_anno;
        }
    }

    for (cur = md->prioritize_annotations; cur; cur=cur->next()) {
        if (cur->pred->equals(predicate->pred->name)) {
            new_anno = new struct corAnnoStruct(cur, 0, predicate->prior_annos);
            predicate->prior_annos = new_anno;
        }
    }
}

/*****************************************************************************/
static void init_module_preds(corQFModuleData *md)
{
    corModulePred *predicates = md->predicates->next;

    for (; predicates; predicates = predicates->next) {
        Name name = predicates->pred->name;
        int arity = predicates->pred->arity;
        predicates->offset = corget_offset(md->F, name, arity, 1);

        switch (predicates->kind) {
        case QueryPredKind:
            predicates->disk_rel =
                diskrel_anno(md, name, predicates->schema_def);
            predicates->check_subsum =
                cordetermine_subsum_status(md, name, md->module_info.CheckSubsum);
            break;

        case LocalPredKind:
            predicates->disk_rel =
                diskrel_anno(md, name, predicates->schema_def);
            if (coris_magic(name)) {
                 predicates->check_subsum = cordetermine_subsum_status(md,name,
                    (md->module_info.CheckSubsum | md->module_info.MultiSet));
            }
            else {
                 predicates->check_subsum = cordetermine_subsum_status(md, name,
                    md->module_info.CheckSubsum);
            }
            break;

        case MagicQueryPredKind:
            predicates->disk_rel = diskrel_anno(md, name,
                predicates->schema_def);
            predicates->check_subsum = cordetermine_subsum_status(md, name,
                (md->module_info.CheckSubsum | md->module_info.MultiSet));
            break;

        case ExternalPredKind:
            break;
        }
        corget_annos(md, predicates);
    }
}

/*****************************************************************************/
/*
    This routine creates the structures within the QFMaterializedModuleData,
    and inserts it into the IterativeMethod table. It assumes that SCC
    analysis has already been performed w.r.t. the particular exported
    predicate.
*/

static void interpret_materialized_qf(corQFMaterializedModuleData *md,
    corExportInfo *export_info, corParserStruct& corparserStruct)
{
    corRuleInfoArray *rArray = md->ruleArray;
    Name exported_name = export_info->pred_name;

    corScc *scc;
    corModulePred *predicates = md->predicates;

    // Make sure that all Relations used by the query are initialized
    // by init_rule_pointers. Use add_pred_extern to enter them into
    // the predicates collection to keep track of them.
    // First enter the the query and magic predicates:

    corPredicate *query_predicate = export_info->predicate;
    coradd_pred_extern(query_predicate, QueryPredKind, predicates);

    int query_arity = query_predicate->arity;
    int magic_arity = export_info->adornment.count();
    char *magic_name = new char[exported_name->length()+4+query_arity];
    corsprint_magic_name(magic_name, exported_name->string(),
        &export_info->adornment);

    corPredicate *magic_predicate = corLookupPredicate(corparserStruct,
        corEnterSymbol(magic_name), magic_arity);
    delete [] magic_name;
    if (magic_predicate) {
        coradd_pred_extern(magic_predicate, MagicQueryPredKind, predicates);
    }
    else {
        fprintf(corexEnv->error_file,
            "Warning: No magic predicate for exported predicate %s\n",
            exported_name->string() );
        if (corparserStruct.CurModule.SaveModule) {
            md->no_magic = 1;
        }
    }

    // Then enter the remaining predicates.
    for (scc = corSccList; scc != NULL; scc = scc->next) {
        corPredicate *pred; corClause *clause; int ipred; corLiteral** literals;

        // Add predicates for non-recursive clauses in this scc.
        for (pred = scc->preds; pred != NULL; pred = pred->next) {
            clause = pred->clauses;
            // If pred->clauses == *pred->clause_tail, then
            // the pred has no clauses, and hence is an ExternalPredKind.
            if (clause != *pred->clause_tail) {
                coradd_pred_extern(pred, LocalPredKind, predicates);
            }
            else if (pred->coris_mp_done()) {
                // NOTE: this is for ordered search
                coradd_pred_extern(pred, LocalPredKind, predicates);
            }
            else
                coradd_pred_extern(pred, ExternalPredKind, predicates);

            // This also determines if clause is recursive.
            for (; clause != *pred->clause_tail; clause = clause->pnext) {
                if (!clause->seen) continue;
                clause->recursive = 0;
                for (literals = clause->preds, ipred = clause->num_preds;
                     --ipred>= 0; literals++)
                {
                    coradd_pred_extern(literals[0]->predicate, ExternalPredKind,
                        predicates);

                    // with Ordered Search, have to special case the 
                    // negated literals
                    if (corparserStruct.CurModule.UseOrdSearch) {
                        if (literals[0]->negated) continue;
                    }
                    // don't need to special case for done_m_p since 
                    // it is covered by the next if condition

                    if (literals[0]->predicate->scc == scc)
                        clause->recursive = 1;
                }
            }
        }
    }

    md->export_pred = export_info->predicate;
    md->module_info = corparserStruct.CurModule;
    if (md->module_info.UsePredicateSN) {
        if (md->module_info.UseOrdSearch) {
            fprintf(corexEnv->error_file,
                "CORAL error: Cannot use PSN with Ordered Search ... yet :-)");
            fprintf(corexEnv->error_file, " using basic SN instead\n");
            md->use_psn = 0;
            md->module_info.UsePredicateSN = 0;
        }
        else if (md->module_info.SaveModule) {
            fprintf(corexEnv->error_file,
                "CORAL error: Cannot use PSN with SaveModule ... yet :-)");
            fprintf(corexEnv->error_file, " using basic SN instead\n");
            md->use_psn = 0;
            md->module_info.UsePredicateSN = 0;
        }
        else if (md->module_info.SupMagicIndexing) {
            fprintf(corexEnv->error_file,
               "CORAL error: Cannot use PSN with SupMagicIndexing ... yet :-)");
            fprintf(corexEnv->error_file, " using basic SN instead\n");
            md->use_psn = 0;
            md->module_info.UsePredicateSN = 0;
        }
    }
    else
        md->use_psn = md->module_info.UsePredicateSN;

    interpret_scc(md, export_info->predicate->scc, export_info,
        corfind_ext_magicquerypred(corparserStruct));

    // move some of the run time work from init_rule_pointers to
    // compile time here. Determine what flags need to be set, etc.
    init_module_preds(md);
    
    // Insert the module structure along with its associated exported
    // query form into the IterativeMethod table.
    if (exported_name) {
        corDerivedMethod* query_method = new corDerivedMethod(exported_name,
            export_info->adornment, cordo_query, (corQFModuleData *)md); 
    }

    //    md->print(stdout);  // DEBUG: added by Ken

       // insert the module to ModuleTable in CurDB: added by Ken
   if (corCurDB->ModuleTable == NULL)
     corCurDB->ModuleTable = corAllocTab(COR_MODULE_TABLE_INCR, NULL);

   Association *ptr = corSymbolLookup(corCurDB->ModuleTable, md->module_info.name);
   ptr->arg = (Name)md->module_info.name;
   ptr->val = md;
   corTabInserted(corCurDB->ModuleTable, ptr);
}

/*****************************************************************************/
static void create_module_table(corParserStruct& corparserStruct,
    corTable* &module_table, corRuleInfoArray* &module_rules)
{
    // Create the Table structure (containing a pointer to each Relation used).
    int pred_count = 0;
    for (corPredicate *pred = corparserStruct.AllPredicates; pred != NULL;
        pred = pred->allprednext)
    {
        pred_count++;
    }

    module_table = new corTable(pred_count);
    module_rules = new corRuleInfoArray;

    // This enters the name and arity of each relation in the Table into the
    // Table array 'module_table'. The relations are also initialized at the
    // same time.

    int i = 0;
    for (pred = corparserStruct.AllPredicates; pred != NULL;
        pred=pred->allprednext, i++)
    {
        module_table->AddValue(i, pred->name, pred->arity);
        pred->frame_offset = i;
    }

    // For each clause, create a rule-info structure that contains the 
    // information necessary to evaluate it.

    int rule_number = 0;
    FOR_EACH_RULE(clause, corparserStruct.rules) {
        clause->rule_number = rule_number++;
        module_rules->AddRule(corcreate_rule_info(module_table, clause));
    } END_EACH_RULE
}

/*****************************************************************************/
/*
 *      Add sup_magic predicate to appropriate Scc. Used in NestedSccEval only
 * Added by Ken
 */
void coradd_sup_to_scc(corParserStruct &corparserStruct)
{ // 
  // traverse the list of predicate, for each sup, find the scc it belongs to
  for (corPredicate *pred = corparserStruct.AllPredicates; pred != NULL;
        pred=pred->allprednext) 
    if (pred->coris_supp()) {
      // get the first predicate in the rhs of clause, e.g. m_p
      corPredicate *first_pred_in_clause = pred->clauses->preds[0]->predicate;
      // get the name of predicate which is in the same Scc as this sup
      // i.e. p
      char *name_of_pred_in_same_scc = first_pred_in_clause->name->string() + 2;
      corPredicate *pred_in_same_scc = corLookupPredicate(corparserStruct, 
                            corEnterSymbol(name_of_pred_in_same_scc), -1);
      pred->scc = pred_in_same_scc->scc;
    }
}

/*****************************************************************************/
/*
 * Make a list of external predicates for each Scc.
 * Used in NestedSccEval only. Added by Ken
 */
void cormake_ext_pred_list(corParserStruct &corparserStruct)
{ // For rules with body predicates that are derived and not in the 
  // same SCC as the head predicate, add those body predicates to the 
  // list of external predicates of the SCC of the head.

  FOR_EACH_RULE(rule, corparserStruct.rules) {
    // check each predicate in the body
    for (int i=0; i < rule->num_preds; ++i) 
      if (!rule->preds[i]->base_pred &&
      rule->preds[i]->predicate->scc != 
      rule->head->predicate->scc) {
    // derived predicate in lower Scc
    // first check if it is already in the list
    int exist = 0;
    for (struct PredList *ext_preds = rule->head->predicate->scc->ext_preds;
         ext_preds != NULL;
         ext_preds = ext_preds->next)
      if (ext_preds->pred == rule->preds[i]->predicate) {
        exist = 1;
        break;
      }
    
    /* if (!exist) {
      // add it to ext_preds of the Scc of the head
      struct PredList *pred_list = 
      pred_list->pred = rule->preds[i]->predicate;
      pred_list->next = rule->head->predicate->scc->ext_preds;
      rule->head->predicate->scc->ext_preds = pred_list;
    }*/
      }
  } END_EACH_RULE   
}

/*****************************************************************************/
/*
 * Top-level routine to interpret a CORAL module. For each exported query
 * form, a QFModuleData structure is created that encapsulates the information
 * needed to evaluate that query. The exported query form and the associated 
 * QFModuleData structure are stored in a table of IterativeMethods.
 */

void corinterpret_module(corParserStruct &corparserStruct)
{

    corExportInfo *export_info;
    corTable *module_table;
    corRuleInfoArray *module_rules;

    if (corparserStruct.CurModule.ExportedPreds == NULL) {
        fprintf(corexEnv->error_file, "CORAL: Error - No exported query form.\n");
        return;
    }

    // Create the Table structure (containing a pointer to each Relation used).
    create_module_table(corparserStruct, module_table, module_rules);

    corQFModuleData *module_data;

    // Process each exported query form independently.
    for (export_info = corparserStruct.CurModule.ExportedPreds;
         export_info != NULL; export_info = export_info->next)
    {
        Name query_symbol;
        Name exported_name = export_info->pred_name;
        int query_arity;

        if (corparserStruct.CurModule.UsePipelining) {
            module_data = new corQFPipelinedModuleData(module_table, module_rules,
                corparserStruct);
        }
        else {
            module_data = new corQFMaterializedModuleData(module_table,
                module_rules, corparserStruct);
        }

        if (export_info->adorn_symbol == 0) {
            query_arity = 0;
            query_symbol = exported_name;
        }
        else {
            Name exported_adorn_name = export_info->adorn_symbol;
            query_arity = (int) exported_adorn_name->length();
                
            if (corparserStruct.CurModule.UsePipelining) {
                int len = (int) exported_name->length() + 1;
                char *query_name = new char[len+1];
                sprintf(query_name,"%s",exported_name->string());
                query_symbol = corEnterSymbol(query_name);
                delete [] query_name;
            }
            else {
                int len = (int) exported_name->length() + 1 + query_arity;
                char *query_name = new char[len+1];
                   corsprint_pred_name(query_name, exported_name->string(),
                    exported_adorn_name->string());
                query_symbol = corEnterSymbol(query_name);
                delete [] query_name;
            }
        }

        // Ensure that the exported query form is a valid one.
        export_info->predicate = corLookupPredicate(corparserStruct, query_symbol,
            query_arity);
        if (!export_info->predicate) {
            fprintf(corexEnv->error_file,
                "Error: Exported predicate %s not defined in module\n",
                query_symbol->string() );
            continue;
        }

        if (corparserStruct.CurModule.UsePipelining) {
            corinterpretPipelined((corQFPipelinedModuleData *) module_data,
                export_info, corparserStruct);
        }
        else {
            // For MATERIALIZED modules
            // Do Scc (Strongly Connected Components) analysis.
            FOR_ALL_RULES(clause) {
                clause->seen = 0;
                clause->recursive = 0;
                for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                    clause->preds[ipred]->predicate->work = NULL;
                    clause->preds[ipred]->predicate->mark = 0;
                }
            } END_EACH_RULE
            corinitialize_scc_structs();

            if (corparserStruct.CurModule.SingleScc ||
                corparserStruct.CurModule.UseOrdSearch ||
                corparserStruct.CurModule.HasPrioritize)
            {
                mark_predicates_invoked_from(export_info->predicate);
                corgroup_predicates_in_one_scc(export_info->predicate);
                clean_up_after_one_scc();
            }
            else {
                mark_predicates_invoked_from(export_info->predicate);
            if (corparserStruct.CurModule.NestedSccEval) {
              corfind_predicates_invoked_from(export_info->predicate, 1);
              coradd_sup_to_scc(corparserStruct);
              cormake_ext_pred_list(corparserStruct);
            }
            else
              corfind_predicates_invoked_from(export_info->predicate);
            }
            interpret_materialized_qf((corQFMaterializedModuleData *)module_data,
                export_info, corparserStruct);
        }
    }
}

/*****************************************************************************/
void corcompile_single_rule(corParserStruct &corparserStruct, corTable* &module_table,
    corRuleInfoArray* &module_rules, corSemiNaive &snrule)
{
    // Create the Table structure (containing a pointer to each Relation used).
    create_module_table(corparserStruct, module_table, module_rules);

    // There is only one RuleInfo structure
    ASSERT((module_rules->numRules == 1));

    // There was only one rule in the parserStruct
    ASSERT((corparserStruct.rule_list->link.next == NULL));

    corClause *clause = corparserStruct.rule_list;

    corcompile_nonrec_rule(snrule, clause, module_rules->rInfoPtrs[0],
        corexEnv->C_non_ground_facts_default);
}

/*****************************************************************************/
void cornumber_sccs(struct corScc *Sccs) // added by Ken
{
  // number the SCCs after the scc analysis is done
  for (int scc_no = 0; Sccs != NULL; scc_no++, Sccs = Sccs->next)
    Sccs->scc_no = scc_no;
}
