/*****************************************************************************/
//    CompileRule.C
/*****************************************************************************/

#include "CompileRule.h"

#include "Literal.h"
#include "Unparse.h"
#include "SymbolTable.h"
#include "Database.h"        // BuiltinDB
#include "RelationMisc.h"
#include "Annotations.h"
#include "Backtrack.h"
#include "Predicate.h"

/*****************************************************************************/
//    WARNING: The following routine damages the varlists that are passed to it.

corVarLink *corintersect_var_lists(corVarLink *varlist1, corVarLink *varlist2)
{
    corVarLink *ptr1, *ptr2;
    corVarLink *head=NULL, *prev=NULL;

    for (ptr1=varlist1; ptr1;) {
        for (ptr2=varlist2; ptr2; ptr2=ptr2->next) {
            if (ptr1->arg->var == ptr2->arg->var)
                break;
        }
        if (ptr2) {
            if (! head) head = ptr1;
            if (! prev)
                prev = ptr1;
            else {
                prev->next = ptr1;
                prev = ptr1;
            }
            ptr1 = ptr1->next;
            prev->next = NULL;
        }
        else ptr1 = ptr1->next;
    }
    return head;
}

/*****************************************************************************/
/*
    Construct a list of all the variables in the bound arguments of the given
    predicate.  Format the list as a Arg * for a predicate.
*/

corVarLink * corbuild_var_list(corLiteral * pred, corVarLink *list)
{
    FOR_EACH_ARG(pred_arg, pred->args) {
        list = pred_arg->var_list(list);
    } END_EACH_ARG

    return list;
}

/*****************************************************************************/
//     WARNING: The following routine damages the varlists that are passed to it.

void corcreate_an_annotation(corPredAnnotations *&annos, corLiteral *lit,
    corVarLink *var_list)
{
    corVarLink *link;
    int var_count = 0;  // count of number of variables in var_list
    int const_count = 0; // count of number of constants in lit->args

    for (link = var_list ; link != NULL; link = link->next)
        var_count++;

    FOR_EACH_ARG(a, lit->args) {
        if (a->hash(NULL) != VarHashValue) const_count++;
    } END_EACH_ARG;

    // no point generating an index with all arguments free !!
    if (!var_count && !const_count) return;

    corArgList *arglist = new corArgList(var_count);

    for (link = var_list; var_count > 0; var_count -- ) {
        corVarLink *next = link->next;
        (*arglist)[var_count-1]= link->arg;
        delete link;
        link = next;
    }
   
    Name name;
    if (lit->kind == COR_MAGIC) {
        char *buf = new char [lit->pred->length()+10];
        corsprint_magic_name(buf,lit->pred->string(),"");
        name = corEnterSymbol(buf);
    }
    else if (lit->kind == COR_DONE_MAGIC) {
        char *buf = new char [lit->pred->length()+15];
        sprintf(buf, "done_");
        char *buf1 = buf + strlen(buf);
        corsprint_magic_name(buf1,lit->pred->string(),"");
        name = corEnterSymbol(buf);
    }
    else {
        name = lit->pred;
    }

    coradd_make_index_annotation(annos,name, &(lit->adorn),&(lit->args),arglist);
}

/*****************************************************************************/
corVarLink *corvars_between(corClause *adrule, int start, int end, corVarLink *rest,
    int *predOrders )
{
    int i;
    corVarLink *result=rest;

    for (i=start; i <= end; i++) {
        if (predOrders)
            result = corbuild_var_list(adrule->preds[predOrders[i]], result);
        else 
            result = corbuild_var_list(adrule->preds[i], result);
    }
    return result;
}

/*****************************************************************************/
int coris_a_builtin(Name name)
{
    int ret_val = 1;
    Association *ptr1;
    ptr1 = corSymbolLookup(corBuiltinDB.RelationTable, name);

    if (HashNone(ptr1->arg)) ret_val = 0;
    //else ret_val = !((BuiltinRelation *)(ptr1->val))->is_printable_builtin();

    return ret_val;
}

/*****************************************************************************/
static int lengthof(corVarLink *vars)
{
    int length = 0;
    corVarLink *ptr1;

    for (ptr1=vars; ptr1; length++)
        ptr1 = ptr1->next;

    return length; 
}

/*****************************************************************************/
/*
    This funtion tries to find the literal in the rule that is best suited to
    be placed immediately after 'start' in the join order.
*/

static int find_best_next_literal(corClause *rulePtr, int *predOrders,
    corVarLink *leading_vars, int start, int end)
{
    int i;
    int best = start;
    int bestval = -1;
    corVarLink *overlap_vars;

    if (coris_a_builtin(rulePtr->preds[predOrders[start]]->pred)) return start;

    for (i = start; i <= end; i++) {
        if (coris_a_builtin(rulePtr->preds[predOrders[i]]->pred)) continue;
        overlap_vars = corintersect_var_lists(
        corbuild_var_list(rulePtr->preds[predOrders[i]],NULL), leading_vars);
        if (lengthof(overlap_vars) > bestval) {
            bestval = lengthof(overlap_vars);
            best = i;
        }
    }

    return best;
}                            

/*****************************************************************************/
void corgenerateIndexes(corSemiNaive *snrule, corPredAnnotations *&annos)
{
    int i, j, k;
    corVarLink *leading_vars, *indexing_vars;

#    define RULEPTR snrule->rInfo->rule    
#    define PREDINDEX(i) snrule->predOrders[i]

    // The following code indexes predicates in the rule
    for (i=0; i < RULEPTR->num_preds; i++) {
        leading_vars = corvars_between(RULEPTR,0,i-1,NULL, snrule->predOrders);
        if (i < snrule->first) {
            j = find_best_next_literal(RULEPTR, snrule->predOrders,
                leading_vars, i, snrule->first);
            k = snrule->predOrders[j];
            while (j > i) {
                snrule->predOrders[j] = snrule->predOrders[j-1];
                j--;
            }
            snrule->predOrders[j] = k;
        }

        // no point building an index on an arithmetic predicate
        if (!coris_a_builtin(RULEPTR->preds[PREDINDEX(i)]->pred)) {
            indexing_vars = corintersect_var_lists(
            corbuild_var_list(RULEPTR->preds[PREDINDEX(i)],NULL), leading_vars);
            corcreate_an_annotation( annos,
                RULEPTR->preds[PREDINDEX(i)], indexing_vars);
        }
    }

    // if there is a grouping argument in the head, create an index on the
    // the head predicate that is bound on all arguments except the grouping
    // argument

    if (corhas_grouping_args(RULEPTR->head->args)) {
        char *buf = new char[RULEPTR->head->args.count() + 1];
        for (i=0; i< RULEPTR->head->args.count(); i++) {
            if (RULEPTR->head->args[i].kindof()==COR_GROUPING)
                buf[i] = 'f';
            else
                buf[i] = 'b';
        }
        buf[i] = '\0';
        corArgList *new_arglist = new corArgList(1);
        (*new_arglist)[0] = corEnterSymbol(buf);

        coradd_make_index_annotation(annos, RULEPTR->head->pred,
            &(RULEPTR->head->adorn), new_arglist, NULL);
    }

#    undef RULEPTR
#    undef PREDINDEX
}

/*****************************************************************************/
void corinitPredOrders(corSemiNaive *sn)
{
    if (sn->rInfo->num_literals) {
        sn->predOrders[0] = sn->first;
        for (int i = 1; i < sn->rInfo->num_literals; i++) {
            if ( i <= sn->first)
                sn->predOrders[i] = i-1;
            else sn->predOrders[i] = i;
        }
     }
}

/*****************************************************************************/
void initFailureBacktrack(corSemiNaive *sn, int ng_facts)
{
    for (int i = sn->rInfo->num_literals-1; i >0; i--) {
        sn->firstFailureBacktrack[i] =
            corgetfirstFailureBacktrackPoint(sn, i, ng_facts);
        sn->nextFailureBacktrack[i] =
            corgetnextFailureBacktrackPoint(sn, i, ng_facts);
    }

    sn->firstFailureBacktrack[0] = -1;
    sn->nextFailureBacktrack[0] = -1;
}

/*****************************************************************************/
corRuleInfo *corcreate_rule_info(corTable *F, corClause *rule)
{
    int num_preds = rule->num_preds;
    int num_deletes = rule->head_deletes_count;
    int ipred;

    // Variable declarations
    
    corRuleInfo *rInfo = new corRuleInfo(rule);

    //  Adding info for rInfo->arg_list, and rInfo->relations.

    //  ArgList **arglist = new ArgList*[num_preds+1];
    
    // NOTE : Should change Arg** to ArgList *
    for (ipred = 0; ipred < num_preds; ipred++) {
        // CHECK
        rInfo->arg_list[ipred] = rule->preds[ipred]->args;
    }

    rInfo->arg_list[num_preds] = rule->head->args;
    for (ipred=0; ipred < num_deletes; ipred++) {
        rInfo->arg_list[num_preds+1+ipred] =
            rule->head_deletes[ipred]->args;
        // CHECK
    }

    // If the predicate is negated, a negative value is placed in the
    // rInfo->relations[] field. This info is used by SemiNaive::evaluate() 
    for (ipred = 0; ipred < num_preds; ipred++) {
        if (rule->preds[ipred]->negated) {
            rInfo->relations[ipred]= -2 - (rInfo->offsets[ipred] =
                corget_offset(F, rule->preds[ipred]));
        }
        else {
            rInfo->relations[ipred]=
                (rInfo->offsets[ipred] = corget_offset(F, rule->preds[ipred]));
        }
    }

    if (rule->head->negated) {
        rInfo->relations[num_preds] = -2 - (rInfo->offsets[num_preds] =
            corget_offset(F, rule->head));
    }
    else {
        rInfo->relations[num_preds] =
            (rInfo->offsets[num_preds] = corget_offset(F, rule->head));
    }

    for (ipred = 0; ipred < num_deletes; ipred++) {
        rInfo->relations[num_preds+1+ipred] =
            -2 - (rInfo->offsets[num_preds+1+ipred] =
            corget_offset(F, rule->head_deletes[ipred]));
    }

    rInfo->relations[num_preds+num_deletes+1] = -1;
    rInfo->offsets[num_preds+num_deletes+1] = -1;

    return(rInfo);
}

/*****************************************************************************/
void corcompile_nonrec_rule(corSemiNaive &sn, corClause *clause, corRuleInfo *rInfo,
    int non_ground)
{
    sn.first = 0;
    sn.recursive = 0;
    sn.Add_Rule_Info(rInfo, clause, non_ground);
    
    for (int jpred = 0; jpred < clause->num_preds; jpred++)
        sn.markOffsets[2*jpred] = sn.markOffsets[2*jpred+1] = -1;
}

/*****************************************************************************/

