// *************************************************************************
//    CoralInit.C
// *************************************************************************

#include <new.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "CoralInit.h"

#include "Globals.h"
#include "Database.h"
#include "Execute.h"
#include "CoralError.h"
#include "CoralBreak.h"
#include "Util.h"
#include "InitBuiltins.h"
#include "DontCareArg.h"
#include "BuiltinSymbols.h"
#include "SMUtil.h"


// *************************************************************************

static void exceptionHandler(int recvd_signal);
static void brokenPipe(int recvd_signal);
static void freeStoreException();

// *************************************************************************

extern void sm_init();

char coral_name[200];        // full pathname of the executing program

int corstartMemUsage;
struct rusage corstartRUsage;
struct rusage corSystemUsage;
int corbeginMem;

//corDatabaseStruct corExodusDB(corEnterSymbol(COR_EXODUS_WS));

extern corArg* corTheParameterArg;

extern "C" {
    int corfulldir(char *path);
    };

// *************************************************************************
//    System specific defines & includes

#ifdef UNIX_HP
    typedef void (*COR_SIGNAL_HANDLER) (int);
#    include <sys/syscall.h>
#    define getrusage(a, corb)  syscall(SYS_GETRUSAGE, a, corb)
#else
#    ifdef __GNUC__
#        define COR_SIGNAL_HANDLER
#    else
#    define COR_SIGNAL_HANDLER SIG_PF
#    endif
#endif

// *************************************************************************
//    Initializes CORAL.  Called from main routine.

void init_coral(char *pgm)
{
#    ifndef DEBUG
        // Catch signals that might otherwise cause termination
        for (int i = SIGINT; i < SIGPIPE; i++) {
            if (i != SIGKILL) {
#                ifdef __GNUC__
                    signal(i, (exceptionHandler)) ;
#                else
                    signal(i, (COR_SIGNAL_HANDLER)(exceptionHandler)) ;
#                endif
            }
        }

          // Catch broken pipes (An Explain that quit)
#       ifdef __GNUC__
            signal(SIGPIPE, (brokenPipe));
#       else
            signal(SIGPIPE, (COR_SIGNAL_HANDLER)(brokenPipe));
#       endif

#    else
        // for debugging purposes, catch only SIGINT (CTRL-C)
#        ifdef __GNUC__
            signal(SIGINT, (exceptionHandler)) ;
#        else
            signal(SIGINT, (COR_SIGNAL_HANDLER)(exceptionHandler)) ;
#        endif
#    endif

    sprintf(coral_pid, "%d", getpid());

    // If neither LocalConnection() nor BindToPort() has been called,
    // set up a local connection with an empty stack.
    if (! corclients->serv_sock)
        corclients->LocalConnection(0);

    // Provide exception handler for new()
    set_new_handler(freeStoreException);

    // create the CurrModuleInfo structure
    corCurrModuleInfo = new corModuleInfo();

    // create the functor table
    corcreateFunctorTable();

    // init builtin symbols
    corinit_builtin_symbols();

    strcpy(coral_name, (pgm ? pgm : "coral"));

    if ((corexEnv->C_exec_mode == COR_DOING_INTERPRET) ||
        (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE))
    {
        // initialize global stats for CPU and memory usage
#        ifndef UNIX_HP
            getrusage(RUSAGE_SELF, &corstartRUsage);
#        endif
        corstartMemUsage = (int)sbrk(0);

#        ifdef WITH_PERSISTENCE
            if (corexEnv->C_connect_exodus_default) {
                // Create exodus workspace
                corCurDB = &corExodusDB;        
                corCurDB->RelationTable = corAllocTab(COR_RELATION_TABLE_INCR, NULL);    
                // Add exodus database to the Database Table
                if (!coraddDBStruct(corCurDB)) {
                    CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add exodus DB",
                        "init-coral", __FILE__, __LINE__);
                }
                else 
                    sm_init();
            }
#        endif

        // Initialize TheParameterArg used for C++ host variables
        corTheParameterArg = &corTheDontCareArg;

        // Remove the stdin.M file
        unlink("stdin.M");

        // Initialize builtin database.
        corBuiltinDB.name = corEnterSymbol("builtin_ws");
        corinitBuiltins();
        corCurDB = &corBuiltinDB;

        // Add builtin database to the Database Table
        if (!coraddDBStruct(corCurDB)) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add builtin DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }

        // Create default database
        corDatabaseStruct *db = corcreateDB(corEnterSymbol(COR_DEFAULT_DB));
        if (!db) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add default DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }
        corCurDB = db ;

        if (corexEnv->C_read_coralrc_default) {
            // Read .coralrc if it exists, else read input from $(HOME)/.coralrc
            if (!coralConfig.server_mode &&
                    !corconsultFile(".coralrc", ".coralrc")) {
                char *home_path = getenv("HOME") ;
                char *tmp_name = new char[strlen(home_path) + 15] ;
                strcpy(tmp_name, home_path);
                strcat(tmp_name, "/.coralrc");
                corconsultFile(tmp_name, tmp_name);
                delete [] tmp_name;
            }
        }
    }
    else {
        corexEnv->C_do_history_default = 0;
    }

#    ifndef UNIX_HP
    getrusage(RUSAGE_SELF, &corSystemUsage);
#    endif
    corbeginMem = (int)sbrk(0);
}

// *************************************************************************
//    Called when quitting coral

void exit_coral()
{
#   ifdef WITH_PERSISTENCE
    if (corexEnv->C_exec_mode == COR_DOING_INTERPRET) {
        shutdown_server();
    }
#   endif

    corclearDatabases();

    //clearSymbols();
    corexEnv->flush_all_outputs();
}

// *************************************************************************
//    Signal handler

static void exceptionHandler(int recvd_signal)
{
    CORAL_error(COR_SIGNAL_RECEIVED,NULL,NULL,__FILE__,__LINE__);
    fprintf(corexEnv->error_file, "Received signal no: %d\n", recvd_signal);

    //    Code for recovering from abnormal situation
    if (recvd_signal != SIGINT) {
        fprintf(corexEnv->error_file, "Cannot recover from received signal\n");
        fprintf(corexEnv->error_file, "Initiating shutdown procedures\n");
#        ifdef WITH_PERSISTENCE
            abortTrans();
#        endif
        exit_coral();
        exit(1);
    }

    if (corexEnv->C_interrupt_raised == 1) {
        fprintf(corexEnv->error_file, "Received second interrupt\n");
        fprintf(corexEnv->error_file, "Initiating shutdown procedures\n");
#        ifdef WITH_PERSISTENCE
            abortTrans();
#        endif
        exit_coral();
        exit(1);
    }

    corexEnv->C_interrupt_raised = 1;
}

//**************************************************************************
//    Signal handler for broken pipe (caused by Explain that disconnects)

static void brokenPipe(int recvd_signal)
{
    // Assume it was the current users explain (pretty safe assumption)

      // Turn off explain and reset associated structures
    corcurUser->corexEnv.explain_fp = (FILE *) NULL;

    corcurUser->corexEnv.GlobalRelationOptions &= ~REL_EXPLAIN;

      // Notify user
    if (corcurUser->corexEnv.error_file)
        fputs("Explain disconnected - turning explain off.\n",
                            corexEnv->error_file);

    if (corcurUser->explain_sock){
        delete corcurUser->explain_sock;
        corcurUser->explain_sock = (CoralSock *) NULL;
    }
}

//**************************************************************************
//    Exception handler for the new() system call

static void freeStoreException()
{
    CORAL_error(COR_NO_MEMORY,NULL,"freeStoreException",__FILE__,__LINE__);
    fflush(corexEnv->print_file);
    fflush(corexEnv->error_file);
    exit(1) ;
}

//**************************************************************************

