/*****************************************************************************/
//    PersistentIndexSet.C
/*****************************************************************************/

#ifdef WITH_PERSISTENCE

#include "PersistentRelation.h"


#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"


#include "PersistentIndexSet.h"

/*****************************************************************************/
int corPersistentIndexSet::insert_index(corGenericIndex* index)
{
    if (!SM_initialized) return COR_I_FAIL;

    corPersistentIndex *p_index = (corPersistentIndex *) index;

    for (int i = 0; i < num_in_array; i++) {
        if (p_index->argset != NULL) {
            //    argument form indexing
            corPersistentIndex *index_i = (corPersistentIndex *)(Indices[i]);
            if (index_i->argset != NULL &&
                *p_index->argset == *index_i->argset)
            {
#                if DO_TRACE
                    if (corexEnv->GlobalRelationOptions & REL_DISPLAY_INDEXOPS) {
                        fprintf(corexEnv->trace_file,
                            "Duplicate index: not inserted\n");
                    }
#                endif
                return(COR_I_FAIL);
            }
        }
    }


    //    new index to be inserted in list
    if (num_in_array == max_in_array) {
#        ifdef DEBUG
            fprintf(corexEnv->error_file, "Too many indices: not inserted\n");
#        endif
        return COR_I_FAIL;
    }
        
    int e;
    
    corFileDescStruct *fd = & (((corPersistentRel *)(p_index->relation))->fd);
    OID *oid = & (((corPersistentRel *)(p_index->relation))->oid);
    USERDESC *u_desc;
    
    //    PRAVEEN:: These error checks are erroneous, since it is possible that
    //    the index does not really need to be created on disk !!

    // Check that an index can be added
    int index_offset;
    if ((index_offset = fd->canAddIndex()) < 0 ) {
        CORAL_error(COR_INDEX_ERR, "Exceeded maximum number of indexes",
            "add_index");
        return COR_I_FAIL;
    }
    
    ASSERT((p_index->argset != NULL) &&
        (p_index->argset->len() <= COR_MAX_ARGLIST_SIZE));

    // Create a new index on the file on the specified attributes
    p_index->chosen_index = fd->addIndex(index_offset, p_index->argset);


    // if a new index has been created
    if (p_index->chosen_index == index_offset) {
        if (OID_IS_INVALID(*oid)) {
            // this means that this relation is a temporary relation that has
            // not been entered into the catalog. 
            Indices[num_in_array++] = (corGenericIndex *) p_index;
            return COR_I_SUCCESS;
         }

        // Change the catalog entry to reflect the new index
        if (!SM.retrieveObject(oid, &u_desc)) {
            CORAL_error(COR_ESM_ERR, NULL, "PersistentIndexSet::insert_index");
            return COR_I_FAIL;
        }

        // Safer to write the entire object, instead of calculating the
        // offset.  You never know with C++ class structures what their
        // true sizes are !!
        e = sm_WriteObject(SM.bufGroup(), 0, sizeof(corFileDescStruct),
            (char *)(fd), u_desc, TRUE );
        if (ErrorCheck(e, "sm_WriteObject")) return COR_I_FAIL;

        Indices[num_in_array++] = (corGenericIndex *) p_index;
        return COR_I_SUCCESS;
    }

    // Otherwise, the index does not have to be created on disk.  All the same,
    // it is not yet recorded in the indexset.  This can happen during startup
    // when indexes are being added to a relation object based on what
    // actually exists on disk.
    Indices[num_in_array++] = (corGenericIndex *) p_index;
    return COR_I_PARTIAL_SUCCESS;
}

/*****************************************************************************/
void corPersistentIndexSet::fill_index(corGenericIndex* index)
{
    SCANDESC *s_desc;
    USERDESC *u_desc;
    BOOL eof;

    corPersistentIndex *p_index = (corPersistentIndex *) index;

    corFileDescStruct& fd = ((corPersistentRel *)(p_index->relation))->fd;

    // If there are any existing tuples in the file, filter them
    // through the index

    int e = sm_OpenScanWithGroup(&(fd.fid), PHYSICAL_SCAN, SM.bufGroup(),
        &s_desc, NULL);
    if (ErrorCheck(e, "sm_OpenScanWithGroup")) return;
    
    int i;
    for (i = 0;;i++) {
        e = sm_ScanNextObject(s_desc, 0, READ_ALL, &u_desc, &eof);
        if (ErrorCheck(e, "sm_ScanNextObject")) break;
        
        // If there are no more tuples, break
        if (eof) break;


        // start index scan for the first tuple :this is being done in this
        // fashion to work around an exodus bug
        if (i == 0) fd.startIndexScan(p_index->chosen_index);

        fd.addObjectToIndex(p_index->chosen_index, u_desc);
    }

    if (i) {
        // end index scan
        fd.endIndexScan(p_index->chosen_index);
    }
    
    e = sm_CloseScan(s_desc);
    ErrorCheck(e, "sm_CloseScan");
        
}

/*****************************************************************************/
//    find an index that best matches the pattern contained arg_list, bind_env

corGenericIndex *corPersistentIndexSet::find_index(corArgList &arg_list,
    corBindEnv *bindenv, int *tryindexp /* = NULL*/)
{
    
    if ( (tryindexp) && ((*tryindexp) >= 0) &&    ((*tryindexp) < num_in_array) )
        return ((corGenericIndex *) Indices[*tryindexp]);

    if (num_in_array == 0) {
        return NULL;
    }

    // find bitvector of pattern 
    int arity = arg_list.count();

    corBitVector adornment(arity);
    int num_bound = 0;            // number of bound arg positions in pattern
    for (int i = 0; i < arity; i++) {
        int bound = 1;
        // check which arguments are ground and set those bits to 1; rest 0
        if (arg_list[i].hash(bindenv) == VarHashValue)
            bound = 0;
        if (bound == 1) num_bound++;
        adornment.set(i, bound);
    }

    // if it is an all-free query, do not choose an index !
    if (!num_bound) return NULL;

    corGenericIndex *return_index = NULL;
    double max_weight = -1.0;
    
    for (i = 0; i < num_in_array; i++) {
        double cur_weight = 0.0;
        if (((corPersistentIndex *) Indices[i])->argset != NULL) {
            if (arity != ((corPersistentIndex *) Indices[i])->argset->len())
                continue;
            if ((num_bound == arity) &&
                (adornment == *(((corPersistentIndex *) Indices[i])->argset)))
            {
                // All arguments are bound - no better match can be found.
                if (tryindexp) *tryindexp = i;
                return ((corGenericIndex *)Indices[i]);
            }
            for (int j = 0; j < arity; j++) {
                if (adornment.test(j) &&
                    ((corPersistentIndex *) Indices[i])->argset->test(j))
                {
                    cur_weight += 1.0;        // both are bound
                    continue;
                }
                if (((corPersistentIndex *) Indices[i])->argset->test(j)) { 
                    //    and !adornment.test(j) the index has more bound
                    //    arguments; it is useless
                    cur_weight = 0.0; 
                    break;
                }
            }
        }
        if (cur_weight > max_weight) {
            // need to update information
            max_weight = cur_weight;
            if (tryindexp) *tryindexp = i;
            return_index = (corGenericIndex *) Indices[i];
        }
    }

    return return_index;
}

/*****************************************************************************/

#endif WITH_PERSISTENCE

