//**************************************************************************
//    ParseStack.C
//
//    Shaun Flisakowski - Oct 1994
//**************************************************************************

#include "ParseStack.h"

#include "Globals.h"
#include "ClientProtocol.h"

//**************************************************************************

extern FILE *yyin;

extern "C" int   yyparse();
extern "C" int   yylex(YYSTYPE *yylval);
extern "C" int   sslex(YYSTYPE *yylval);

extern YY_BUFFER_STATE yy_create_buffer( FILE *file, int size );
extern YY_BUFFER_STATE ss_create_buffer( FILE *file, int size );
extern void            yy_switch_to_buffer( YY_BUFFER_STATE buff );
extern void            ss_switch_to_buffer( YY_BUFFER_STATE buff );
extern void            yy_delete_buffer( YY_BUFFER_STATE buff );
extern void            ss_delete_buffer( YY_BUFFER_STATE buff );

//**************************************************************************
corParseStack::~corParseStack()
{
      // Need to allow items on stack to call destructors
    while(!is_empty())
        pop();
}

//**************************************************************************
// ParseStack::parse:
// This function parses the top parse_cnt items on the stack (default of 1),
// if the parse_cnt is given as 0, all items on the stack are parsed.
// If a syntax error causes yyparse to exit, the err_mess is printed
// to the error file pointer, err_fp. The error message defaults to
// "Parse Error.\n".  The number of items actually parsed is returned.
//
//**************************************************************************
int  corParseStack::parse(FILE *err_fp, int parse_cnt, char *err_mess)
{
    if (is_empty()){
        corexEnv->c_module_separator = 1;
        return(0);
    }

    for (int i=0; i<parse_cnt || !parse_cnt; i++) {

        while(yyparse())
            fputs(err_mess,err_fp);

        if (is_empty()){
            corexEnv->c_module_separator = 1;
            break;
        }
    }

    if (is_empty())
        i++;

    return(i);
}

//**************************************************************************
// ParseStack::push:
// This function places a file pointer at the top of the parse stack,
// it does NOT prepare the file to be parsed, which requires a call to 
// reset_postion.  This call is publicly available to allow multiple items
// to be placed on the stack without the overhead of called reset_position
// multiple times.  If you are only placing one item on the stack, you
// should probably be using the handle_new_file function instead.
//
//**************************************************************************

void corParseStack::push(FILE *fp1, int line_num, char *fname)
{
    corParserEnv *tmp = new corParserEnv;

    tmp->fp = fp1;
    tmp->ptype = FromFile;
    tmp->lexer = yylex;
    tmp->yybuff = yy_create_buffer(fp1, YYBUFF_SIZE);

    tmp->next = top;

    if (fname) {
        tmp->filename = new char[strlen(fname)+1];
        strcpy(tmp->filename, fname);
    } else
        tmp->filename = NULL;

    tmp->cur_line_num = line_num;
    if (!top)
        bottom = tmp;
    top = tmp;
}

//**************************************************************************
// ParseStack::push:
// This function places a character buffer at the top of the parse stack,
// it does NOT prepare the buffer to be parsed, which requires a call to 
// reset_postion.  This call is publicly available to allow multiple items
// to be placed on the stack without the overhead of called reset_position
// multiple times.  If you are only placing one item on the stack, you
// should probably be using the handle_new_buff function instead.
//
//**************************************************************************

void corParseStack::push(char *parsebuff, int line_num, char *bname)
{
    corParserEnv *tmp = new corParserEnv;
    tmp->ptype = FromBuff;

    tmp->parsebuff = parsebuff;
    tmp->parselimit = parsebuff + strlen(parsebuff);
    tmp->parseptr = parsebuff;

    tmp->lexer = sslex;
    tmp->yybuff = ss_create_buffer((FILE *) NULL, YYBUFF_SIZE);

    if (bname) {
        tmp->filename = new char[strlen(bname)+1];
        strcpy(tmp->filename, bname);
    } else
        tmp->filename = NULL;

    tmp->next = top;
    if (!top)
        bottom = tmp;
    top = tmp;
}

//**************************************************************************
// ParseStack::put_on_bottom:
// This function places a file at the bottom of the parse stack,
// it does NOT prepare the file to be parsed, which requires a call to 
// reset_postion.  This call is publicly available to allow items to be
// placed on the bottom of stack, verus the top, which occurs when using
// handle_new_file.  This is useful, for example, when consulting a
// list of files provided by the user, you can place the items on the
// bottom of the (previously empty?) stack, then when you start parsing,
// you will be parsing in the order the files were listed. 
//
//**************************************************************************

void corParseStack::put_on_bottom(FILE *fp1, int line_num, char *fname)
{
    corParserEnv *tmp = new corParserEnv;
    tmp->fp = fp1;
    tmp->ptype = FromFile;
    tmp->lexer = yylex;
    tmp->yybuff = yy_create_buffer(fp1, YYBUFF_SIZE);

    tmp->next = (corParserEnv *) NULL;

    if (fname) {
        tmp->filename = new char[strlen(fname)+1];
        strcpy(tmp->filename, fname);
    } else
        tmp->filename = NULL;

    tmp->cur_line_num = line_num;

    if (bottom)
        bottom->next = tmp;
    if (!top)
        top = tmp;
    bottom = tmp;
}

//**************************************************************************
// ParseStack::pop:
// This function removes the topmost ParserEnv from the ParseStack. This 
// ParserEnv could have been for a file or a character buffer.  This call
// does NOT prepare the next file to be parsed, which requires a call to 
// reset_postion.  If you want the next file prepared for parsing you
// probably should be using the abort_file() or get_next_file() functions.
//
//**************************************************************************

void corParseStack::pop()
{
    if (!top)
        return;

    if ((top_of_stack())->ptype == FromFile)
        yy_delete_buffer(top->yybuff);
    else
        ss_delete_buffer(top->yybuff);

    corParserEnv *tmp = top->next;
    if (top->filename)
        delete [] top->filename;
    delete top;
    if (!tmp)
        bottom = (corParserEnv *) NULL;

    top = tmp;
}

//**************************************************************************
// ParseStack::handle_new_file:
// This function places a file pointer at the top of the parse stack, the
// next call to yyparse() would then parse it, taking whatever action it
// caused.  
//
//**************************************************************************

void corParseStack::handle_new_file(FILE *fp, char *fname)
{
    push(fp,1, fname);
    reset_position();
}

//**************************************************************************
// ParseStack::handle_new_buff:
// This function places a character buffer at the top of the parse stack,
// the next call to yyparse() would then parse it, taking whatever action it
// caused.  
//
//**************************************************************************

void corParseStack::handle_new_buff(char *buff, char *bname)
{
    push(buff,1,bname);
    reset_position();
}

//**************************************************************************
// ParseStack::reset_position:
// This function is for interal use, after a push, pop, or put-on-bottom
// to the parse stack, this function is called to reset parseEnv, the
// pointer to the current ParserEnv (the one at the top of the stack)
//
//**************************************************************************

void corParseStack::reset_position()
{
    if ( !(corparseEnv = top_of_stack()))
        return;

    if (corparseEnv->ptype == FromFile){

        yy_switch_to_buffer(corparseEnv->yybuff);
        yyin = top_file();
        corparseEnv->show_prompt = corparseEnv->incr_command_count =
            (yyin == corcurConn->conn_sock->in);

    } else {        // ptype == FromBuff
        ss_switch_to_buffer(corparseEnv->yybuff);
        corparseEnv->show_prompt = corparseEnv->incr_command_count = 0;
    }
}

//**************************************************************************
// ParseStack::get_next_file:
// This function takes an item off of the parse stack, and prepares the
// the stack for the next call to yyparse(), it is used in yywrap() when
// EOF is reached so that multiple items on the stack can be parsed by
// simply calling yyparse() repeatedly. 
//
// *************************************************************************

int corParseStack::get_next_file()
{
    if (is_empty())
        return(0);

    if ((top_of_stack())->ptype == FromFile)
        fclose(yyin);

    pop();
    reset_position();

    if (is_empty())
        return(0);

    return(1);
}

//**************************************************************************
// ParseStack::abort_file:
// This function aborts the file (or string) that is currently being parsed,
// the next file (or string) on the stack is prepared for parsing so that
// yyparse() can simply be called again.  This function is called by the
// parser when it has problems, or when an interrupt has been recieved.  
//
//**************************************************************************

void corParseStack::abort_file()
{
    if (!is_empty()){

        if (((top_of_stack())->ptype == FromFile) && (top_file() != stdin))
            fclose(yyin);

        pop();
        reset_position();
    }
}

//**************************************************************************
// ParseStack::show_stack:
// This function shows all the files that are currently on the stack.
// It is for debugging only, and so prints directly to stderr.
//
//**************************************************************************

void corParseStack::show_stack()
{
    corParserEnv *tmp;

    fputs("-----Showing Stack: \n",stderr);
    if (top){

        fputs("Starting at top (current item being Parsed).\n",stderr);
        for (tmp=top; tmp; tmp=tmp->next){
            if (tmp->filename)
                fprintf(stderr, "%s:\t", tmp->filename);
            else
                fputs("--no name--:\t", stderr);
            fputs((tmp->ptype == FromFile) ? "File\n" : "Buffer\n",stderr);
        }

    } else
        fputs("Stack is Empty.\n",stderr);

    fputs("-----Done Showing Stack \n",stderr);
}
//**************************************************************************
