/*****************************************************************************/
//    Context.C
/*****************************************************************************/

#include "Context.h"

#include "Globals.h"
#include "Unify.h"
#include "StorageRelation.h"

/*****************************************************************************/
corContext::corContext()
{
    first = NULL;
    last = NULL;
    first_marked = NULL;
    last_marked = NULL;
}

/*****************************************************************************/
void corContext::add_new_goal(corGoalNode *g_node, corGoalNode *par)
{
    // NOTE: this does not do the duplicate checking on the Context.
    // that HAS to be done elsewhere

    g_node->create_goalnode_info(par);
    corContextNode *c_node = new corContextNode(g_node, this);

    // the new goal and context are unmarked

    // first, we handle the extreme case of seed facts on Context
    // NOTE: all unmarked seed facts have a node_number of 0

    if (par == NULL) {
        // insert at end of Context
        insert_last(c_node);
        return;
    }

    //    With a non-null "par" field, have to insert it just before
    //    the next marked context node from context node of par
    //    or at end of context if context node of par is last marked node

    ASSERT((par->is_marked()));
    corContextNode *par_c_node = par->find_context_node();
    ASSERT((par_c_node != NULL));
    ASSERT((par_c_node->is_marked()));

    // node number of newly inserted node is same as that of its parent's node
    c_node->node_number = par_c_node->node_number;

    corContextNode *par_next_c_node = par_c_node->next_marked;

    if (par_next_c_node == NULL) { // insert at end of Context
        insert_last(c_node);
    }
    else {
        insert_before(c_node, par_next_c_node);
    }

}

/*****************************************************************************/
void corContext::insert_last(corContextNode *c_node)
{
    if (first == NULL) {
        // this is the first context node in Context
        first = c_node;
        last = c_node;
    }
    else { 
        c_node->prev = last;
        last->next = c_node;
        last = c_node;
    }
}

/*****************************************************************************/
void corContext::insert_before(corContextNode *c_node, corContextNode *par_next_c_node)
{
    c_node->next = par_next_c_node;
    c_node->prev = par_next_c_node->prev;
    par_next_c_node->prev = c_node;
    c_node->prev->next = c_node;
}

/*****************************************************************************/
/*
    NOTE: deleting goal nodes from a context node in the non-ground case
    can only be done by setting a delete flag.  how does that affect the
    running time?

    In the all ground case, any node which is not a singleton node
    should have no unmarked goals in it.  so for the ground case,
    we have to worry about deletion only from singleton nodes,
    which should not mess anything up.
*/

void corContext::delete_goal(corGoalNode *g)
{
    //    NOTE: ordered search needs to delete only unmarked goals during
    //    subsumption checks

    ASSERT((!(g->is_marked())));

    corContextNode *g_c_node = g->find_context_node();
    ASSERT((g_c_node != NULL));

    //    the following is a hack
    if ((g_c_node->first_marked_goal == NULL) &&
        (g_c_node->first_unmarked_goal == NULL))
    {
        // something shady happened
        return;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_ordsearch) {
            fprintf(corexEnv->trace_file, "\n\tTrying to delete GoalNode:  \n");
            g->print_all(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\tCorresponding ContextNode is:  \n");
            g_c_node->print_all(corexEnv->trace_file);
        }
#    endif

    if (g_c_node->is_marked()) {
        // unmarked goal in marked node
        // mark goal as deleted, do nothing else
        g->do_delete();
        g->c_node = NULL;
        g->parent_goal = NULL;
        g->parent_in_set = NULL;

        // NOTE: this is safe since unmarked goal nodes in marked
        //       context nodes are guaranteed to be leaf nodes because
        //       of the set manipulation operations
    }
    else {
        // unmarked goal in singleton unmarked node -- can delete everything
        ASSERT((g_c_node->first_marked_goal == NULL));
        ASSERT((g_c_node->first_unmarked_goal == g_c_node->last_unmarked_goal));

        // marking the goal itself as deleted
        g->do_delete();
        g->c_node = NULL;
        g->parent_goal = NULL;
        g->parent_in_set = NULL;

        // unlink the context node
        unlink_node(g_c_node);

        g_c_node->free_all();
        delete g_c_node;
    }         
}

/*****************************************************************************/
void corContext::unlink_node(corContextNode *g_c_node)
{
    if (g_c_node->next == NULL) {
        // last node in Context
        if (g_c_node->prev != NULL) {
            // not first node in Context
            g_c_node->prev->next = NULL;
        }
        last = g_c_node->prev;
    }
    else {
        // not last node in Context
        g_c_node->next->prev = g_c_node->prev;
        if (g_c_node->prev != NULL) {
            // not first node in Context
            g_c_node->prev->next = g_c_node->next;
        }
    }

    if (g_c_node->is_marked()) {
        if (g_c_node->next_marked == NULL) {
            // last marked node
            if (g_c_node->prev_marked != NULL) {
                // not first marked node
                g_c_node->prev_marked->next_marked = NULL;
            }
            last_marked = g_c_node->prev_marked;
        }
        else {
            // not last marked node
            g_c_node->next_marked->prev_marked = g_c_node->prev_marked;
            if (g_c_node->prev_marked != NULL) {
                // not first marked node
                g_c_node->prev_marked->next_marked = g_c_node->next_marked;
            }
        }
    }

    // update context info
    if (last == NULL) first = NULL;
    if (last_marked == NULL) first_marked = NULL;
}

/*****************************************************************************/
corGoalNode *corContext::get_unmarked_goal(int &any_new_fact_in_done)
{
 any_new_fact_in_done = 0;

 if (last == NULL) {
   // empty context
   if (corexEnv->dbg_ordsearch) {
     fprintf(corexEnv->trace_file, "        OrdSearch: Empty context.\n");
    }
   return NULL;
  }

 corGoalNode *fac;
 corContextNode *c = last; 
 fac = c->get_unmarked_goal();
 if (fac == NULL) {
   // node contains no unmarked subgoal
   ASSERT((c->is_marked()));
   if (corexEnv->dbg_ordsearch) {
     fprintf(corexEnv->trace_file,
         "OrdSearch: moving goals to done.\n"); 
    }
   c->move_goals_to_done();
   any_new_fact_in_done = 1;
   unlink_node(c);

   c->free_all();
   //    ERROR!!  This causes major memory overwrite problems.
   //    If c is deleted here, there are still pointers to it from the
   //    c_node of some GoalNode.  This gets accessed later via
   //    find_context_node, and we get screwed. : PRAVEEN
   //delete c;
   // since we have moved facts to the done relation new inferences
   // can possibly be made in Ordered Search
  }
 else {
   // have an unmarked goal
   if (corexEnv->dbg_ordsearch) {
     fprintf(corexEnv->trace_file, "OrdSearch: found unmarked goal: ");
     fprintf(corexEnv->trace_file,
         "    %s(", corSymbolString(fac->owner_rel->name));
     fac->args().print(NULL,corexEnv->trace_file);
     fprintf(corexEnv->trace_file, ") \n");
     print_all(corexEnv->trace_file);
    }

   if (!(c->is_marked())) {
     c->mark_node();
     update_marked_info(c);
     
     // set node number of newly marked context node
     if (c->prev_marked == NULL) {
       c->node_number = 1;
      }
     else {
       c->node_number = c->prev_marked->node_number + 1;
      }
    }
  }
 return fac;
}

/*****************************************************************************/
void corContext::update_marked_info(corContextNode *c_node)
{
    // we are assuming that only the highest unmarked node on
    // Context ever gets marked, which is fine for ordered search

    if (first_marked == NULL) {
        // no marked nodes in Context yet
        first_marked = c_node;
        last_marked = c_node;
    }
    else {
        // only marked context nodes are linked up in this way
        c_node->prev_marked = last_marked;
        c_node->next_marked = NULL;
        last_marked->next_marked = c_node;
        last_marked = c_node;
    }
}

/*****************************************************************************/
void corContext::collapse_nodes(corGoalNode* from, corGoalNode* par, corGoalNode* to)
{
    if (from == to) return;        // no collapsing required
    
    ASSERT((from->is_marked()));
    ASSERT((par->is_marked()));
    ASSERT((!(to->is_marked())));

#    ifdef DO_TRACE
        if (corexEnv->dbg_ordsearch) {
            fprintf(corexEnv->trace_file, "OrdSearch: collapsing nodes\n");
            fprintf(corexEnv->trace_file, "           from:  ");
            from->args().print(from->bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\n           par :  ");
            par->args().print(par->bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\n           to  :  ");
            to->args().print(to->bindenv, corexEnv->trace_file);
        }
#    endif 

    corContextNode *from_c_node = from->find_context_node();
    corContextNode *par_c_node = par->find_context_node();

    if (from_c_node != par_c_node) {
        // collapse marked nodes into from_c_node
        corContextNode *c_tmp;
        corContextNode *pnxt_tmp = par_c_node->next_marked;

        for (c_tmp = from_c_node->next_marked; c_tmp && c_tmp != pnxt_tmp; ) {
            // First check if any cycle of negative dependencies exist.
            // Only the c_tmp nodes need to be checked, not the 
            // from_c_node, since that could be negatively dependent on
            // a previous goal, without any problems

            // ASSERT((!(c_tmp->negated)));
            if (c_tmp->negated) {
                fprintf(corexEnv->error_file,
                    "Cycle of negative dependencies detected!\n");
            }

            from_c_node->merge_nodes(c_tmp);
            unlink_node(c_tmp);

            //    NOTE: cannot call c_tmp->free_all() since the goal node 
            //    links have also been linked up TMP!!
            delete c_tmp;
            c_tmp = from_c_node->next_marked;
        }
    }
    
    // now collapse "to" into "from"
    corContextNode *to_c_node = to->find_context_node();

    // ASSERT((!(to_c_node->negated)));
    if (to_c_node->negated) {
        fprintf(corexEnv->error_file, "Cycle of negative dependencies detected!\n");
    }

    if (!(corsubsumes_tuple(from, to, corexEnv->dbg_ordsearch))) {
        from_c_node->merge_nodes(to_c_node);        
        unlink_node(to_c_node);

        //    NOTE: cannot call to_c_node->free_all() since the goal node
        //    links have also been linked up TMP!!
        delete to_c_node;
    }
    else {
        // no merging takes place
        to->do_delete();  // to make sure get_next_tuple does not get this
        unlink_node(to_c_node);
        to_c_node->free_all();
        // TMP!!
        delete to_c_node;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_ordsearch) {
            print_all(corexEnv->trace_file);
        }
#    endif
}


/*****************************************************************************/
void corContext::print_all(FILE *file) 
{
    ASSERT((last->next == NULL)); 
    fprintf(file, "OrdSearch: printing state of Context.\n"); 
    fprintf(file, "\t---------------------------\n");
    for (corContextNode *temp = first; temp != NULL; temp = temp->next) {
        temp->print_all(file);
    }
    fprintf(file, "\t---------------------------\n");
}

/*****************************************************************************/
void corContext::free_all()
{
    // should also go through the context nodes and free all of those 
    first = NULL;
    last = NULL;
    first_marked = NULL;
    last_marked = NULL;
}

/*****************************************************************************/

