/*****************************************************************************/
//    GoalNode.C
/*****************************************************************************/

#include "GoalNode.h"

#include "RelationMisc.h"
#include "Util.h"
#include "StorageRelation.h"
#include "Execute.h"

/*****************************************************************************/
corGoalNode::corGoalNode(corStorageRelation *rel, int neg) : corTuple()
{
    set_defaults(rel, neg);
}

/*****************************************************************************/
corGoalNode::corGoalNode(int arity, corStorageRelation *rel, int neg): corTuple(arity)
{
    set_defaults(rel, neg);
}

/*****************************************************************************/
corGoalNode::corGoalNode(corArgList *arglist, corBindEnv *env, corStorageRelation *rel,
    int neg)
    : corTuple(arglist, env)
{
    set_defaults(rel, neg);
}

/*****************************************************************************/
corGoalNode::corGoalNode(corArgList&a, corStorageRelation *rel, int neg) : corTuple(a)
{
    set_defaults(rel, neg);
}

/*****************************************************************************/
corGoalNode::corGoalNode(corArgList *arglist, corStorageRelation *rel, int neg)
    : corTuple(arglist)
{
    set_defaults(rel, neg);
}

/*****************************************************************************/
void corGoalNode::set_defaults(corStorageRelation *rel, int neg)
{
    mark = 0;
    parent_goal = NULL;
    c_node = NULL;
    owner_rel = rel;

    negated = neg;

    // now set the fields related to makeset() of union-find
    parent_in_set = this;
    rank_in_set = 0;

#    ifdef DEBUG
        corC_NewTupleSize += sizeof(corGoalNode) - sizeof(corTuple);
#    endif
}

/*****************************************************************************/
corGoalNode *corGoalNode::find_canonical_elem ()
{
    // this is the find() operation in tarjan's union-find operation
    // it automatically does path compression, while doing a single
    // pass through the parent_in_set chain

    if (this != parent_in_set) {
        parent_in_set = parent_in_set->find_canonical_elem();
    }

    return parent_in_set;
}

/*****************************************************************************/
corContextNode *corGoalNode::find_context_node()
{
    // GoalNode maintains a c_node field, but for reasons of 
    // asymptotics, only the c_node field in the canonical GoalNode
    // of the set is the CORRECT one.

    corGoalNode *can_g_node = find_canonical_elem();

    return can_g_node->c_node;
}

/*****************************************************************************/
/*
    NOTE: unmarked goal nodes in marked context nodes better be 
    guaranteed to be leaf nodes because this assumption is used elsewhere
*/
/*****************************************************************************/
corGoalNode *corGoalNode::link_goals(corGoalNode *g)
{
    // this is the link() operation of tarjan's union-find
    if (rank_in_set < g->rank_in_set) {
        parent_in_set = g;
        g->c_node = c_node; // since g is the root of merged set
        return g;
    }
    if (rank_in_set = g->rank_in_set) {
        // going to make "this" as g's parent 
        rank_in_set++;
    }
    g->parent_in_set = this;
    g->c_node = NULL;                // for safety only
    return this;
}

/*****************************************************************************/
int corGoalNode::is_magic_goal()
{
    return (owner_rel->r_kind == COR_R_MAGIC);
}

/*****************************************************************************/
corStorageRelation *corGoalNode::find_done_relation()
{
    ASSERT((is_magic_goal()));
    if (!(owner_rel->mp_done_reln)) {
        char *rel_str = owner_rel->name->string();
        int rel_str_len = strlen(rel_str);

        char *done_str = cormake_done_prefix();
        int done_str_len = strlen(done_str);

        char *rel_done_str = new char [rel_str_len+done_str_len+1];

        strcpy(rel_done_str, done_str);
        strcpy(rel_done_str+done_str_len, rel_str);
        rel_done_str[done_str_len+rel_str_len] = '\0';
        
        Name rel_done_name = corEnterSymbol(rel_done_str);
        owner_rel->mp_done_reln = (corStorageRelation *)
            corfind_local_relation(rel_done_name, owner_rel->arity());
            // this only searches on the current frame

        if (!owner_rel->mp_done_reln) {
            // not on frame
            owner_rel->mp_done_reln = (corStorageRelation *)
                cormake_local_relation(rel_done_name, owner_rel->arity(), 0);
                // NOTE: where can this relation be deleted???
        }

        delete rel_done_str;        // free up some space here
    }
    return owner_rel->mp_done_reln;
}

/*****************************************************************************/
void corGoalNode::print_all(FILE *file)
{
    fprintf(file, "\t\tPrinting GoalNode information.\n");
    fprintf(file, "\t\tMark = %d;    ", mark);
    fprintf(file, "%s (", corSymbolString(owner_rel->name));
    corTuple::printon(file);
    fprintf(file, ")\n");
}

/*****************************************************************************/

