/*****************************************************************************/
//    LinkedRelation.C
/*****************************************************************************/

#include "LinkedRelation.h"

#include "Globals.h"
#include "RelationList.h"

/*****************************************************************************/
corLinkedRelation::corLinkedRelation(int arity, int delta)
    : corStorageRelation(arity, delta), index_set(this) 
{
    //    It is essential to have at least one index - a lot of code depends
    //    on this.
    
    corBitVector* bv1 = new corBitVector(arity);

    for (int i = 0; i < arity; i++) {
        // all arguments bound
        bv1->set(i);
    }
    corHashIndex* h1 = new corHashIndex( this, bv1) ;
    index_set.insert_index( h1 );
}

/*****************************************************************************/
corRMark * corLinkedRelation::getMark()
{
    return new corRMark(lastptr);
}

/*****************************************************************************/
void corLinkedRelation::initialize_iterator( corTupleIterator& iterator )
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,"\nLinkedRelation::initialize_iterator: ");
            iterator.relation->name->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ( ");
            iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " )\n");
        }
#    endif
    iterator.stack_mark.get_mark();

    corTerm t1;
    int all_free = 1;

    for (int i = 0; i < _arity; i++) {
        if (iterator.arg_list[i]->hash(iterator.bindenv) != VarHashValue) {
            all_free = 0;
            break;
        }
    }

    if (!delta_indexed) {
        if (iterator.start_mark) {
            // don't need to go to the hash table
            iterator.ipos = Scanning_New;
            iterator.ppos = (iterator.start_mark)->lastptr;
        }
        else {
            // need to scan from hash table
            // find the appropriate index to be used
            if (all_free) { 
                iterator.ppos = &chain;
                iterator.index = NULL;
            }
            else 
                iterator.index = index_set.find_index(iterator);
        }
    } 
    else {
        if (!iterator.in_delta()) {
            // Starting on main relation
            if (iterator.start_mark) {
                // go to the delta relations
                iterator.ipos = Scanning_Init;
                iterator.delta_rel_pos =  (iterator.start_mark)->rel_p;
            }
            else {
                // need to scan from hash table
                // find the appropriate index to be used
                if (all_free) {
                    iterator.ppos = &chain;
                    iterator.index = NULL;
                }
                else 
                    iterator.index = index_set.find_index(iterator);
            }
        }
        else {
            // In a delta relation
            if (all_free) {
                iterator.ppos = &chain;
                iterator.index = NULL;
            }
            else 
                iterator.index = index_set.find_index(iterator);
        }
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next && !iterator.index) {
            fprintf(corexEnv->trace_file, " : using tuple list !\n");
        }
#    endif
}

/*****************************************************************************/
#define FAIL_RETURN \
{ \
    iterator.ipos = Scanning_Done;     \
    iterator.set_no_match();    \
    return NULL;            \
}

corTuple *corLinkedRelation::get_next_tuple(corTupleIterator& iterator)
{
#    ifdef DO_PROFILE
        if (TableStackCount > 0) {
            ((corTableStack[TableStackCount-1])->moduledata->
                module_info.corGlobalProfileInfo.getnexts)++;
        }
        else {
            (corGlobalProfileInfo.getnexts)++ ;
        }
#    endif
    
    corBindEnv *result_env = iterator.bindenv;
    iterator.tuple_env = NULL;
    
    corTuple *result = NULL;

    if (corexEnv->C_interrupt_raised) FAIL_RETURN;

    // initialize
    if (iterator.ipos == Scanning_Init)
        initialize_iterator(iterator);
    else
        iterator.stack_mark.pop_to();


    if (!iterator.index) {
        // an unbound scan using the linked rel !

        // The current position is remembered in (TupleLink**)iterator.ppos.
        corTupleLink **next_ptr = (corTupleLink**)iterator.ppos;
        corStackMark stack_mark;

        for (;;) {
            if (*next_ptr == NULL) { 
                iterator.stack_mark.pop_to();
                FAIL_RETURN;
            }
            if (iterator.end_mark) {
                if ((iterator.end_mark)->lastptr == next_ptr) { 
                  iterator.stack_mark.pop_to();
                  FAIL_RETURN;
                }
            }
            iterator.ipos++;
            corTuple *tuple = (*next_ptr)->tuple;
            if (tuple->is_deleted()) { 
                // the tuple has been logically deleted
                
                *(next_ptr) = (*next_ptr)->next;
                if (scan_count == 1)
                    delete (*next_ptr);
            
#                ifdef DO_PROFILE
                    if (TableStackCount > 0) {
                        (corTableStack[TableStackCount-1]->moduledata->
                            module_info.corGlobalProfileInfo.deleted_tuples)++ ;
                    }
                    else {
                        (corGlobalProfileInfo.deleted_tuples)++ ;
                    }
#                endif
                continue;
            }
            
            next_ptr = &(*next_ptr)->next;
            iterator.ppos = next_ptr;
            
            if (corunify_literal(iterator, tuple, 0))
                return tuple;
            
            stack_mark.pop_to();
        }
    }
    else {
        //    use the chosen hash-index : ASSUME DELTA_INDEXING

        if (!iterator.in_delta()) {
            // Scanning from main hash table
            result = iterator.index->get_next_tuple(iterator);
            if (result != NULL) return result;
            if (!delta_relations) FAIL_RETURN;
            
            // now go into the delta relations
            // Scanning_Init for delta reln
            iterator.ipos = Scanning_Init;
            iterator.delta_rel_pos = delta_relations->reln_chain;
        }

        // delta_rel_pos has to have been set by this point.
        if (delta_relations != NULL) { 
            // iterator scanning deltas, this == main relation
            for(;;) {
                // Loop till successful delta is found
                if (iterator.ipos == Scanning_Init) {
                    if (iterator.delta_rel_pos == NULL) {
                        FAIL_RETURN;
                    }
                    if (iterator.end_mark && (iterator.delta_rel_pos ==
                        (iterator.end_mark)->rel_p))
                    {
                        FAIL_RETURN;
                    }
                }
                result = iterator.delta_rel_pos->reln->get_next_tuple(iterator);
                if (result != NULL) return result;

                // Done with current delta relation.  Got to go to next delta.
                iterator.delta_rel_pos = iterator.delta_rel_pos->next;
                // To start on new delta reln.
                iterator.ipos = Scanning_Init;      
            }
        }
        else {
            // iterator scanning deltas, this == delta relation
            return iterator.index->get_next_tuple( iterator);
        }
    }
}

#undef FAIL_RETURN

/*****************************************************************************/
void corLinkedRelation::release(corTupleIterator&)
{
    //    Do nothing, since TupleIterator does not maintain extra info to
    //    be released in the case of LinkedRelations.
}

/*****************************************************************************/
corBindEnv *corLinkedRelation::get_next(corTupleIterator& iterator)
{
    corTuple *result = get_next_tuple(iterator);
    if (!result) return NULL;
    return iterator.bindenv;
}

/*****************************************************************************/
int corLinkedRelation::is_subsumed(corTuple *newtuple, corRMark *start_mark,
    corRMark *end_mark)
{
    corTupleLink **next;

    if (start_mark == NULL)
        next = &chain;
    else
        next = (start_mark)->lastptr;

    corStackMark stack_mark;

    for (;;) {
        if (*next == NULL) { 
            return is_agg_subsumed(newtuple, start_mark, end_mark);
        }
        if (end_mark) {
            if ((end_mark)->lastptr == next) { 
                return is_agg_subsumed(newtuple, start_mark, end_mark); 
            }
        }
        corTuple *tuple = (*next)->tuple;
        next = &(*next)->next;

        if (tuple->is_deleted()) {
            // tuple has been logically deleted
            *(next) = (*next)->next;
            if (scan_count == 1)
                delete (*next);
#            ifdef DO_PROFILE
                if (TableStackCount > 0) {
                    (corTableStack[TableStackCount-1]->moduledata->
                        module_info.corGlobalProfileInfo.deleted_tuples)++ ;
                }
                else {
                    (corGlobalProfileInfo.deleted_tuples)++ ;
                }
#            endif
            continue;
        }
            
        if (corsubsumes_tuple(tuple, newtuple, (DISPLAY_INSERTIONS)))
            return 1;

        stack_mark.pop_to();
    }
}

/*****************************************************************************/
int corLinkedRelation::insert_raw(corTuple *tuple)
{
    //    tuple has been logically deleted
    if (tuple->is_deleted()) return 0;

    tlist_insert_tuple(tuple);

    //    inserts tuple into each index constructed for the relation.
      
    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->insert_tuple(tuple);
    }
    return 1;
}

/*****************************************************************************/
void corLinkedRelation::empty_relation(int deleteTuples)
{
    // Empty out relation, but do NOT delete tuples (which may be shared)

    corTupleLink* next;

    for (corTupleLink *link = chain; link != NULL; link = next) {
        if (deleteTuples) delete link->tuple;
        next = link->next;
        delete link;
    }
    chain = NULL;
    lastptr = &chain;
}

/*****************************************************************************/
void corLinkedRelation::print_name( FILE *outf)
{
    fprintf(outf, "%15s/%d  : (base:list)    ", corSymbolString(name), arity());
    index_set.print_adornments(outf);
    fprintf(outf, ": count %d\n", count);
}

/*****************************************************************************/

