// *************************************************************************
//    StorageRelation.C
//**************************************************************************

#include "StorageRelation.h"

#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "RelationList.h"
#include "RelationMisc.h"
#include "BuiltinSymbols.h"
#include "HashSimpleRelation.h"
#include "Unify.h"
#include "DontCareArg.h"
#include "GroupingArg.h"

//**************************************************************************

// temporary hack to avoid changing generic-index.h
// also see persistent-index.C ::PRAVEEN
#define COR_I_PARTIAL_SUCCESS  2

// *************************************************************************
corStorageRelation *corAllocateRelation(Name name, int arity, int delta_indexed)
{
    corStorageRelation *rel;
    rel = new corHashSimpleRelation(arity, delta_indexed);
    rel->name = name;
    rel->r_kind = corfind_r_kind(name);

    return rel;
}

// *************************************************************************
corStorageRelation::corStorageRelation() : corRelation(), corTupleList(), headMark()
{
    char namep[100];
    sprintf(namep,"Rel_%d", corRelCount++);
    name = corEnterSymbol(namep);
    _arity = 0;

    local_options = 0;
    count = 0;
    scan_count = 0;
    ref_count = 1;
    single_success = 0;

    check_subsum = corexEnv->C_check_subsum_default;

    for_supp_index = NULL;

    r_kind = COR_R_ANSWER;  // default

    delta_indexed = corexEnv->C_index_deltas_default;
    delta_relations = NULL;
    free_reln_list  = NULL;

    // initialize circular list of marks
    headMark.next = &headMark;
    headMark.prev = &headMark;

    // to allow reuse of marks
    free_marks = NULL;
}

// *************************************************************************
corStorageRelation::corStorageRelation(int new_arity, int delta)
    : corRelation(new_arity, delta), corTupleList(), headMark()
{
    char namep[100];

    sprintf(namep,"Rel_%d", corRelCount++);
    name = corEnterSymbol(namep);
    _arity = new_arity;

    local_options = 0;
    count = 0;
    scan_count = 0;
    ref_count = 1;
    single_success = 0;
    check_subsum = corexEnv->C_check_subsum_default;
    agg_sel_info = NULL;
    priority_info = NULL;

    _hash = UnknownHashValue;
    _eq_id = UnknownHashValue;

    for_supp_index = NULL;

    ordsearch = 0;
    for_ordsearch = NULL;
    mp_done_reln = NULL;

    if (delta == -1)
        delta_indexed = corexEnv->C_index_deltas_default;
    else
        delta_indexed = delta;
    delta_relations = NULL;
    free_reln_list    = NULL;

    // initialize circular list of marks
    headMark.next = &headMark;
    headMark.prev = &headMark;

    free_marks = NULL;    /* to allow reuse of marks*/
}


// *************************************************************************
int corStorageRelation::insert_tuple(corTuple *tuple)
{
    count++;
    insert_raw(tuple);
    return 1;
}


// *************************************************************************
int corStorageRelation::insert_raw(corTuple *)
{
    CORAL_error(COR_BAD_TUPLE_INSERT,name->string(),"insert_raw");
    return -1;
}


// *************************************************************************
/*
    for now, lets assume that any delta relation that are indexed are 
    hashsimplerelations
*/

corRMark * corStorageRelation::getMark()
{
    corRMark *mark;
    if (!delta_indexed) {
        if (free_marks) {
            mark = free_marks;
            free_marks = free_marks->next;
            mark->init(lastptr);
        }
        else mark = new corRMark(lastptr);
    }
    else {
        corHashSimpleRelation *newreln;
        int create_relation = 1;
        
        if (!delta_relations) {
            delta_relations = new corRelationList();
        }
        else {
            /*
            NOTE: Restore this after adding counts to all relations.
             if (delta_relations->reln_lastptr && 
             delta_relations->reln_lastptr->reln->count == 0) 
             then { // there's no need to create a new reln. 
             // Just do mark->init(delta_relations->reln_lastptr);
                         create_relation = 0;
             }
            */
        }
        
        if (create_relation) {
            if (!free_reln_list) {
                newreln = new corHashSimpleRelation(arity(), delta_indexed,
                    corexEnv->C_delta_buckets_default);

                //    NOTE: This is to circumvent a problem with the hashing
                //    code, which uses the relation name unnecessarily.
                newreln->name = name;
                add_all_indices(newreln);
                delta_relations->append_relation(newreln);
            }
            else {
                // Reuse relation in list
                corRelationLink *new_reln_link = free_reln_list;
                free_reln_list = free_reln_list->next;
                new_reln_link->next = NULL;
                delta_relations->append(new_reln_link);
            }
        }
        if (free_marks) {
            mark = free_marks;
            free_marks = free_marks->next;
            mark->init(delta_relations->reln_lastptr);
        }
        else {
            mark = new corRMark(delta_relations->reln_lastptr);
        }
    }

    // Link it in
    mark->next = &headMark;
    mark->prev = headMark.prev;
    headMark.prev->next = mark;
    headMark.prev = mark;
    return mark;
}

// *************************************************************************
void corStorageRelation::freeMark(corRMark *mark, int)
{
    if (mark == NULL) return;
    delete mark;
}

// *************************************************************************
void corStorageRelation::free_allMarks()
{
    fprintf(corexEnv->trace_file,
        "Relation::free_allMarks() should not be called.\n");
}

// *************************************************************************
// to be used internally, not by user

int corStorageRelation::add_index(corBitVector *bv)
{
    corGenericIndexSet *index_set;
    corGenericIndex *h;
    int ret_val;

    if ((index_set = get_index_set()) == NULL) {
      //fprintf(exEnv->error_file,
      //"StorageRelation::add_index found NULL index_set\n");
      // socket_rel has no index. -wX
      return COR_I_FAIL;
    }
    if ((h    = create_index_object(bv)) == NULL) {
        fprintf(corexEnv->error_file,
            "StorageRelation::add_index unable to create index object\n");
        return COR_I_FAIL;
    }

    if ((ret_val = index_set->insert_index(h)) == COR_I_FAIL) {
        // temporary hack : need to add reclaim_index_object method !!
        // fprintf(exEnv->error_file,
        //            "StorageRelation::add_index unable to insert index\n");
        correclaim_index(h, r_kind);
        return COR_I_FAIL;
    }
    else if (ret_val == COR_I_PARTIAL_SUCCESS) {
        return COR_I_PARTIAL_SUCCESS;
    }

    index_set->fill_index(h);

    // add the same index to the relation pointed to by for_supp_index
    if (for_supp_index != NULL) {
        for_supp_index->add_index(bv);
    }
    
    // add the same index to each of the delta_indexed relations
    if (delta_indexed && delta_relations) {
        for (corRelationLink *tmp=delta_relations->reln_chain; tmp != NULL;
            tmp=tmp->next)
        {
            tmp->reln->add_index(bv);        
        }
    }
    return COR_I_SUCCESS;
}


// *************************************************************************
int corStorageRelation::add_index(char *adorn)
{
    ASSERT(adorn != NULL);

    // MEMORY LEAK : Praveen
    corBitVector *bv = new corBitVector(arity());

    char *adorn1 = adorn;

    for (int i = 0; i < arity(); i++, adorn++) {
        if (*adorn == 0) {
            fprintf(corexEnv->error_file,
                "StorageRelation::add_index adorn too short: \n");
            fprintf(corexEnv->error_file,
                "    --- arity of %s=%d, adorn len=%d)\n", name->string(),
                arity(), i); 
            return(COR_I_FAIL);
        }
        if (*adorn == 'b' || *adorn == '1')
            bv->set(i);
    }
    if (*adorn != 0) {
        fprintf(corexEnv->error_file,"StorageRelation::add_index adorn too long\n");
        fprintf(corexEnv->error_file,
            "    --- arity of %s=%d, adorn len=%d adorn = %s)\n",
            name->string(), arity(), i, adorn1); 
        return(COR_I_FAIL);
    }

    return (this->add_index(bv));
}

// *************************************************************************
int corStorageRelation::add_index(corArgList *args)
{
    ASSERT(args->count() != 0);
    if (args->count() > 1) {
        fprintf(corexEnv->error_file, "More than one argument to add_index ");
        fprintf(corexEnv->error_file, "-- using first one to build index.\n");
    }
    ASSERT((*args)[0].kindof() == COR_SYMBOL_CONST);
    char *arg_str = ((*args)[0].SymbolP())->string();
    return add_index(arg_str);
}

// *************************************************************************
int corStorageRelation::add_index(corArgList *pat, int n_var, corArgList *v_nm)
{
    char *argindex = corsimplify_index(pat, n_var, v_nm);

    if (argindex) 
        return add_index(argindex);

    corGenericIndexSet *index_set;
    corGenericIndex *h;

    if ((h  = create_index_object(pat, n_var, v_nm)) == NULL)
        return COR_I_FAIL;

    if ((index_set = get_index_set()) == NULL)
        return COR_I_FAIL;

    if (index_set->insert_index(h) == COR_I_FAIL)
        return COR_I_FAIL;

    index_set->fill_index(h);
        
    // add the same index to the relation pointed to by for_supp_index
    if (for_supp_index != NULL) {
        for_supp_index->add_index(pat, n_var,v_nm);
    }
    
    // add the same index to each of the delta_indexed relations
    if (delta_indexed && delta_relations) {
        for (corRelationLink *tmp=delta_relations->reln_chain; tmp != NULL;
            tmp=tmp->next)
        {
            tmp->reln->add_index(pat, n_var, v_nm);
        }
    }
    return COR_I_SUCCESS;
}


// *************************************************************************
/*
    Called to add all the indices in one relation on to another (usually
    on to a newly created delta relation. 
*/

int corStorageRelation::add_all_indices(corStorageRelation *reln)
{
    corGenericIndexSet *index_set;
    if (index_set = this->get_index_set())
        return index_set->add_all_indices(reln);
    else
        return 0;
}

// *************************************************************************
/*
    Called to add all the tuples in (this) into reln.
*/

int corStorageRelation::add_all_tuples(corStorageRelation *reln,
    int marks_into_account, int check_dupl, int hide_em)
{
    corGenericIndexSet *index_set;
    
    if (index_set = this->get_index_set()) {
        return index_set->add_all_tuples(reln, marks_into_account,
            check_dupl, hide_em);
    }
    else
        return(0);
}

// *************************************************************************
/*
    Called to convert all tuples in (this) into goal nodes and
    add them to reln and context.  For OrdSearch
*/

int corStorageRelation::add_all_goals(corStorageRelation *reln, corContext *context,
    corStorageRelation *owner_rel)
{
    corGenericIndexSet *index_set;
    
    if (index_set = this->get_index_set()) {
        return index_set->add_all_goals(reln, context, owner_rel);
    }
    else
        return(0);
}

// *************************************************************************
void corStorageRelation::add_agg_sel_info(corAggSelInfo *newinfo) 
{
    // first check that this aggsel has not already been specified
    if (corduplicate_agg_sel_anno(this, newinfo)) return;

    newinfo->next = agg_sel_info;
    agg_sel_info = newinfo;
}

// *************************************************************************
void corStorageRelation::add_agg_sel_anno(corPredAnnotations *anno) 
{
/*
    AggSelInfo *newinfo = new AggSelInfo;
    ArgList *arglist = newinfo->arglist = anno->arglist1;
*/

    if (anno->op_arg.kindof() != COR_FUNCTOR) {
        corannotation_error(anno, "Bad aggregation type");
        return;
    }
    corFuncArg *fop_arg =  anno->op_arg.FuncArgP();

    if (fop_arg->arity() != 1 && fop_arg->functor() != corAnyOpSymbol) {
        corannotation_error(anno,
            "Bad number of arguments for aggregate selection");
        return;
    }

    for (int opargnum=0; opargnum < fop_arg->arity(); opargnum++) {
        corAggSelInfo *newinfo = new corAggSelInfo;

        if (fop_arg->functor() == corMaxOpSymbol)
            newinfo->agg_kind = MaxAggregation;
        else if (fop_arg->functor() == corMinOpSymbol)
          newinfo->agg_kind = MinAggregation;
        else if (fop_arg->functor() == corAnyOpSymbol)
            newinfo->agg_kind = AnyAggregation;
        else {
            // possibly extend to irrelevant fact annotation
            corannotation_error(anno, "Only min, max and any aggregation handled");
            return;
        }

        corArgList *arglist = newinfo->arglist = anno->arglist1;
        newinfo->next = NULL;
        newinfo->indexnum = -1;

        newinfo->agg_value = fop_arg->args[opargnum];
        if (newinfo->agg_value.kindof() != COR_VARIABLE) {
            corannotation_error(anno, "Aggregated argument should be a variable");
            return;
        }
        newinfo->var_count = arglist->max_vars();

        corArgList* copy_args = new corArgList(arglist->arity());
        corVarLink* groupby_vars = anno->arglist2->var_list(NULL);

        corBitVector* groupargs = new corBitVector(arglist->arity());
        groupargs->clear();

        for (int i=0; i < arglist->arity(); i++) {
            if ( (*arglist)[i].kindof() != COR_VARIABLE ) {
                corannotation_error(anno, "Arglist should contain only variables");
                return;
            }
            corVarArg *oldvar = (*arglist)[i].VarArgP();

            // Search for var in anno->arglist2
            if (groupby_vars->contains(oldvar) ) {
                (*copy_args)[i] = oldvar;
                groupargs->set(i);
            }
            else {
                (*copy_args)[i] = make_var( newinfo->var_count++ );
            }

            if (oldvar->var == (newinfo->agg_value.VarArgP())->var)
                newinfo->gb_agg_value = (*copy_args)[i];
        }

        newinfo->groupbylist = copy_args;
        newinfo->bindenv = new corArrayBindEnv(newinfo->var_count);

        //    Add the aggregate selection, and create an index too
        //    WARNING: Better to move the index generation into the rewriting
        //    phase, so that it is visible.

        add_agg_sel_info(newinfo);
        add_index(groupargs);
    }
}


// *************************************************************************
/*
    insert_new:  returns 1 if a tuple was inserted into the end of a relation.
    In the case of aggregation this function calls insert_grouping, and the
    convention  is more complicated - 
        a) returns 1 if the new tuple is inserted
        b) returns 0 in the case that an existing tuple was modified, but
            new tuple not inserted.
*/

int corStorageRelation::insert(corTuple *tuple)
{
#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            fprintf(corexEnv->trace_file, "[Inserting %s(", corSymbolString(name));
            tuple->args().print(tuple->bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")");
        }
#    endif

    if (tuple->arity() != arity()) {
        fprintf(corexEnv->trace_file,
            "Relation::insert - attempt to insert tuple of arity %d \n",
            tuple->arity());
        fprintf(corexEnv->trace_file,
            "                 - in relation %s of arity %d \n",
            name->string(), arity());
        return 0;
    }
    
    if (corhas_grouping_args(tuple->args())) 
        return insert_grouping(tuple->args(),tuple->bindenv);

    if (tuple->bindenv == NULL) {
        tuple->env_size = tuple->args().max_vars();
        return insert_tuple(tuple);
    }
    else {
        corArgList *copy_args = new corArgList(arity());
        //    CopyArgs also renames variables so that they are distinct,
        //    and the bindenv need not be used to distinguish variables.
        int env_size = corCopyArgs(copy_args, & tuple->args(),
                    tuple->bindenv, tuple->arity(), &corUnusedEnv);

#        ifdef DO_TRACE
            if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                if (! DISPLAY_INSERTIONS) {
                    fprintf(corexEnv->trace_file,
                        "[Inserting %s(", corSymbolString(name));
                    copy_args->print(NULL, corexEnv->trace_file);
                    fprintf(corexEnv->trace_file, ")");
                }
                if (env_size == 0) {
                    fprintf(corexEnv->trace_file, " - done (ground)]\n");
                }
                else {
                    fprintf(corexEnv->trace_file,
                        " - done (env-size:%d)]\n", env_size);
                }
            }
#        endif
        corTuple *tuple1 = new corTuple(copy_args);
        tuple1->env_size = env_size;
        //count++;
        insert_tuple(tuple1);
        return 1;
    }
}

// *************************************************************************
int corStorageRelation::insert_grouping(corArgList& args, corBindEnv* env, corBindEnv *,
    corTuple *, int)
{
    corStackMark stack_mark0;

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            fprintf(corexEnv->trace_file,
                "[Insert grouping: %s(", corSymbolString(name));
            args.print(env, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")");
        }
#    endif

    corCArg grouping_arg;
    int grouping_arg_num;
    int i;

    corArgList* temp_args = new corArgList(args.count());

    //    WARNING::  ReturnUnify optimizations needed here!!!
    //    See how this is affected by and will affect return unify optimizations

    for (i=0;i<args.count(); i++) {
        (*temp_args)[i] = args[i];
        if (args[i].kindof() == COR_GROUPING) {
            grouping_arg_num=i;
            grouping_arg = args[i];
            // **** WARNING::
            //     The following code is dangerous - it 
            //     changes a static rule structure.
            //   Make sure the structure is reset before
            //     this function returns.
            //   Since the function can never recursively
            //   call the rule to which the arg belongs,
            //   this modification is safe.

            args[i]= &corTheDontCareArg; 
            // Replace DontCareArg by a new variable !!?????
            // Tricky replacement - unification on args will affect 
            //   grouping arg in case it shares variables with 
            //   the non-grouping arg.
        }
    }

    corTupleIterator iter(this, args, env);
    corStackMark stack_mark1;
    int do_insert = 1; 
    int stat=0;

    for (;;) {
        corStackMark stackmark2;
        corTuple *tuple1 = iter.get_next_tuple();
        stackmark2.pop_to(); 
        // Undo any bindings done by get_next_tuple();
        // WARNING: will work ONLY if tuples don't have bindenvs!!
        // WARNING: the current version of subsumes_args also has this
        //        limitation!!
        if (tuple1==NULL) 
            break;
        if (corsubsumes_arg_lists(args, tuple1->args(), env, iter.tuple_env)) {
#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS) {
                    fprintf(corexEnv->trace_file,
                        " - non-grouping args subsume - (");
                    tuple1->printon(corexEnv->trace_file);
                    fprintf(corexEnv->trace_file, ")]\n");
                }
#            endif
#            ifdef DO_EXPLAIN
                char *copy_arglist;
                if (this->local_options & REL_EXPLAIN) {
                    copy_arglist = new char[5000];
                    int  pos = 0;
                    (tuple1->args()).sprint(copy_arglist, &pos);
                }
#            endif

            do_insert = 0;
            stat = corupdate_grouping(this, tuple1, temp_args, env);

#            ifdef DO_EXPLAIN
                if (this->local_options & REL_EXPLAIN) {
                    corgroup_dump(env, tuple1->_args, this, copy_arglist);
                    delete [] copy_arglist;
                }
#            endif

        }
        stack_mark1.pop_to();
        break;
    }
    stack_mark0.pop_to();

    for (i=0;i<args.count(); i++) {
        args[i] = (*temp_args)[i];
    }

    if (! do_insert) {
#        ifdef DO_TRACE
            if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                if (!DISPLAY_INSERTIONS) {
                    fprintf(corexEnv->trace_file,
                        "[Inserting Grouping %s(", corSymbolString(name));
                    args.print(env, corexEnv->trace_file);
                }
                fprintf(corexEnv->trace_file,
                    ") - no new fact created (old one may have been modified)\n");
            }
#        endif
        return 0;
        // Examine carefully to see if non-ground
        // cases are handled properly.
    }

    corArgList *copy_args = new corArgList(arity());

    //    CopyArgs also renames variables so that they are distinct,
    //    and the bindenv need not be used to distinguish variables.
    //    Grouping::simplify does the ``right'' thing with aggregation - 
    //    it removes the grouping junk and returns the argument.
    //    In fact, for count it just returns a default arg since it 
    //    will never be looked at.
    int env_size = corCopyArgs(copy_args, & args, env, arity(),
        &corUnusedEnv);

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
            fprintf(corexEnv->trace_file,
                "[Inserting Grouping %s(", corSymbolString(name));
            copy_args->print(NULL, corexEnv->trace_file);
            if (env_size == 0)
                fprintf(corexEnv->trace_file, " - done (ground)]\n");
            else
                fprintf(corexEnv->trace_file, " - done (env-size:%d)]\n", env_size);
        }
#    endif

    // the new tuple created has copied the arglist.  subsequent
    // modifications to the tuple by Grouping::update_tuple 
    // can hence update the arglist in place without worrying
    // about it being shared by anyone else

    corTuple *tuple2 = new corTuple(copy_args);
    tuple2->env_size = env_size;
    insert_tuple(tuple2);

#    ifdef DO_EXPLAIN
        if (this->local_options & REL_EXPLAIN)
            corgroup_dump(env, tuple2->_args, this);
#    endif

    return 1;
}

// *************************************************************************
int corStorageRelation::is_subsumed(corTuple *tuple, corRMark *startmark, corRMark *endmark)
{
    corStackMark stack_mark1;

    //    Check for ordinary subsumption
    corTupleIterator iter(this, *(tuple->_args), tuple->bindenv, startmark,
        endmark);
    for (;;) {
        corStackMark stackmark2;
        corTuple *tuple1 = iter.get_next_tuple();
        stackmark2.pop_to(); 
        // Undo any bindings done by get_next_tuple();
        if (tuple1 == NULL) break;
        if (corsubsumes_arg_lists(tuple1->args(), *(tuple->_args), iter.tuple_env,
            tuple->bindenv))
        {
#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS) {
                    fprintf(corexEnv->trace_file, " - subsumed by (");
                    tuple1->printon(corexEnv->trace_file);
                    fprintf(corexEnv->trace_file, ")]\n");
                }
#            endif
            stack_mark1.pop_to();
            return 1;
        }
        stack_mark1.pop_to();
    }
    stack_mark1.pop_to();

    return is_agg_subsumed(tuple, startmark, endmark);
}

// *************************************************************************

#define GETBINDENV(aggsel,env) \
{                            \
    env = aggsel->bindenv;                               \
    aggsel->bindenv = NULL;  /*To allow reentrant use of aggsel->bindenv*/ \
    if(env==NULL) env = new corArrayBindEnv(aggsel->var_count);           \
}

#define RELEASE(aggsel,env) \
{                    \
    if (aggsel->bindenv==NULL) aggsel->bindenv = env;    \
    else delete env;                    \
} 

// *************************************************************************
int corStorageRelation::is_agg_subsumed(corTuple *tuple, corRMark *startmark,
    corRMark *endmark)
{
    //    Check aggregate selections

    //    WARNING::  ReturnUnify optimizations needed here!!!

    corStackMark stack_mark1;
    corAggSelInfo *aggsel;
     
    for (aggsel=agg_sel_info; aggsel != NULL; aggsel=aggsel->next) {
        // For each agg sel in the list
        corStackMark stack_mark2;
        corBindEnv *aggsel_env;
        GETBINDENV(aggsel, aggsel_env);

        corunify_arg_lists(*(aggsel->arglist), *(tuple->_args), aggsel_env, 
            tuple->bindenv);
        corTupleIterator iter(this, *(aggsel->groupbylist), aggsel_env,
            startmark, endmark, &(aggsel->indexnum));

        for (;;) {
            corStackMark stackmark3;
            corTuple *tuple1 = iter.get_next_tuple();

            // aggsel->groupbylist and aggsel->gb_agg_value correspond to
            // the fetched tuple.
            stackmark3.pop_to(); 

            if (iter.no_match() || tuple1==NULL) 
            break;

            if (corsubsumes_arg_lists(*(aggsel->groupbylist), tuple1->args(),
                aggsel_env, iter.tuple_env))
            {
                // Check if new value (aggsel->agg_value) subsumes old
                // value (aggsel->gb_agg_value).
                int stat = corsubsumes_value(aggsel->agg_kind, aggsel->agg_value,
                aggsel->gb_agg_value, aggsel_env, aggsel_env);
                if (stat > 0) {
#                    ifdef DO_TRACE
                        if (corexEnv->dbg_aggregates || DISPLAY_INSERTIONS) {
                              fprintf(corexEnv->trace_file,
                                "\n\t\t - aggsel: new fact subsumes");
                              fprintf(corexEnv->trace_file, ": %s(", name->string());
                              tuple1->printon(corexEnv->trace_file);
                              fprintf(corexEnv->trace_file, ")\n");
                        }
#                    endif
                    delete_id_tuple(this, tuple1);
                }
                else if (stat < 0) {
#                    ifdef DO_TRACE
                        if (corexEnv->dbg_aggregates || DISPLAY_INSERTIONS) {
                              fprintf(corexEnv->trace_file,
                                  "\n\t\t - aggsel: old fact subsumed by");
                              fprintf(corexEnv->trace_file, ": %s(", name->string());
                              tuple1->printon(corexEnv->trace_file);
                              fprintf(corexEnv->trace_file, ")\n");
                        }
#                    endif
                    stack_mark1.pop_to();
                    RELEASE(aggsel,aggsel_env)
                    return 1;
                }
                // New fact not subsumed
                //  using current aggregate selection
            }
            stackmark3.pop_to();
         }
        stack_mark2.pop_to();
        RELEASE(aggsel,aggsel_env)
    }
    stack_mark1.pop_to();

    //    Fact is not subsumed
    return 0;
}

// *************************************************************************
//  The following has to be fixed some time or the other, but is complicated 
//  by structure sharing.
// *************************************************************************
void corStorageRelation::free_all()
{
}

// *************************************************************************
void corStorageRelation::free_indices()
{
}

// *************************************************************************
corStorageRelation::~corStorageRelation()
{
    free_all();
}

// *************************************************************************
int corStorageRelation::insert_in_context(corGoalNode *g_node, corGoalNode *parent_goal)
{
    // first insert tuple in the for_ordsearch relation to index context
    for_ordsearch->insert_tuple(g_node);

    ASSERT((corT_Stack.count != 0));
    ASSERT((corT_Stack.top()->context != NULL));

    corT_Stack.top()->context->add_new_goal(g_node, parent_goal);

    if (check_subsum != 1) return 1;

    corContextNode *g_c_node = g_node->find_context_node();
    ASSERT((g_c_node != NULL));
    int g_node_number = g_c_node->node_number;
    corGoalNode *lowest_marked = NULL;
    int lowest_marked_number = g_node_number + 1;
    
    corArrayBindEnv *g_node_env = NULL;
    if (g_node->bindenv == NULL) {
        // this is a hack because of some restrictions in 
        // subsumes_arg_lists
        g_node_env = new corArrayBindEnv(g_node->env_size);
    }
    else
        g_node_env = (corArrayBindEnv *) g_node->bindenv;

    // now the code to find matching tuples from for_ordsearch relation
    // for subsumption purposes
    corStackMark stack_mark10;
    corTupleIterator iter10(for_ordsearch, g_node->args(), g_node_env);

    for (;;) {
        corStackMark stackmark20;
        corTuple *tuple20 = iter10.get_next_tuple();
        stackmark20.pop_to();
        // Undo any bindings done by get_next_tuple();
        if (iter10.no_match()) break;

        ASSERT((tuple20 != NULL));
        if (tuple20 == g_node) continue; 
        // this can happen because g_node has already been 
        // inserted in for_ordsearch relation 

        corContextNode *tuple20_c_node = 
            ((corGoalNode *) tuple20)->find_context_node();
        ASSERT((tuple20_c_node != NULL));
        int tuple20_node_number = tuple20_c_node->node_number;
        if (corsubsumes_arg_lists(tuple20->args(), g_node->args(),
            iter10.tuple_env, g_node->bindenv))
        {
            // we know that tuple20 subsumes g_node
            stackmark20.pop_to();
            // undo bindings done by subsumes_arg_lists
        
            if (g_node_number < tuple20_node_number) {
                // can delete the new inserted goal node 
                corT_Stack.top()->context->delete_goal(g_node);
                stack_mark10.pop_to();
                return 0;     // not inserted
            }
        }
        stackmark20.pop_to();

        if (corsubsumes_arg_lists(g_node->args(), tuple20->args(),
            g_node_env, iter10.tuple_env))
        {
            stackmark20.pop_to();
                // Undo bindings done by subsumes_arg_lists

            if ((tuple20_node_number <= g_node_number) && 
                    (!(((corGoalNode *) tuple20)->is_marked())) )
            {
                // tuple20 is below g_node on context and can be deleted
                corT_Stack.top()->context->delete_goal((corGoalNode *) tuple20);

                // NOTE: might need to delete goal from the 
                // for_ordsearch relation 
                continue;
            }
            else if ((tuple20_node_number <= g_node_number) &&
                (((corGoalNode *) tuple20)->is_marked()))
            {
                // have to find the lowest marked node subsumed by g_node
                if (tuple20_node_number < lowest_marked_number) {
                    lowest_marked_number = tuple20_node_number;
                    lowest_marked = (corGoalNode *) tuple20;
                }
                continue;
            }
        }
        stackmark20.pop_to();
    }
    stack_mark10.pop_to();

    // have removed all lower unmarked goals subsumed by g_node and
    // found the lowest marked goal subsumed by g_node

    if (lowest_marked_number <= g_node_number) {
        // g_node subsumes something marked below it
        // hence, nodes need to be collapsed

        ASSERT((lowest_marked != NULL));
        corT_Stack.top()->context->collapse_nodes(lowest_marked, parent_goal,
            g_node);
    }

    if (g_node->bindenv == NULL) {
        delete g_node_env;
    }
    return 1; // has been added to context but not subsumed
}

// *************************************************************************
