/*****************************************************************************/
//    Collection.h
/*****************************************************************************/

#ifndef _Collection_H_
#define _Collection_H_

#include <stdio.h>

#include "SymbolTable.h"
#include "Link.h"
#include "Rule.h"

/*****************************************************************************/
/*
    Declarations for a structure to handle collections of objects
    in an easy fashion. It is used to handle the collection of rules
    generated by the parser while parsing the rules in a module.
*/
/*****************************************************************************/

/*****************************************************************************/
/*
    Macros for iterating over collections of rules:
    Example usage:
        FOR_EACH_RULE(rule, ruleset) {
            do_someting(rule);
        } END_EACH_RULE
    The curly braces are optional, but otherwise tools which
    do program indentation (e.g. Emacs C-mode) do the wrong thing.
    Note: 'Clause *rule' is implicitly declared by the macros.
    'ruleset' is a (pointer to) a 'struct collection' of rules.
*/

/*****************************************************************************/

#define FOR_EACH_RULE(RULE, RULESET)\
  { corClause *RULE = (corClause*)(RULESET)->chain;\
    for ( ; RULE; RULE = (corClause*)RULE->link.next) {

/*****************************************************************************/
//    Iterate over query+rules 

#define FOR_ALL_RULES(RULE)\
  { corClause *RULE = corparserStruct.rule_list;\
    for ( ; RULE; RULE = (corClause*)RULE->link.next) {

/*****************************************************************************/
//    Iterate over RULESET (which must be hashable), selecting only rules
//    whose head match teh name PRED.

#define FOR_EACH_RULE_FOR(RULE, PRED, RULESET)\
  { corClause *RULE; Name NAME = PRED;\
    Association *PTR = corSymbolLookup((RULESET)->tab, NAME);\
    if (HashNone(PTR->arg)) RULE = NULL;\
    else RULE = (corClause*)PTR->val;\
    for (;RULE && RULE->link.name==NAME; RULE = (corClause*)RULE->link.next){

#define END_EACH_RULE }}

/*****************************************************************************/
/*
    The following four #define's are because FOR_EACH_RULE, etc. are being
    misused to store collections of predicates rather than rules.
    NOTE: this is a hack, and should be cleaned up to properly use
    predicate structures rather than rule structures.
*/

#define FOR_EACH_PRED FOR_EACH_RULE
#define FOR_ALL_PREDS FOR_ALL_RULES
#define FOR_EACH_PRED_FOR FOR_EACH_RULE_FOR
#define END_EACH_PRED END_EACH_RULE

/*****************************************************************************/

void corinit_collection(struct corcollection*, int);
void coradd_member(Link *member, struct corcollection *corcollection);
Link * corpop_member(struct corcollection *corcollection);
void coradd_rule(struct corClause*, struct corcollection*);

/*****************************************************************************/
/*
    A collection is a set (list) of struct that contain 'Link' fields.
    An object can be in at most one collection.
    A collection is linked in a linear list (using the 'next' field of
    each link, and head by the 'chain' field of the collection).
    A collection can optionally be hashed, in which case
    all objects with the same link->name are consequtive in the list.
    Objects (in a hashed colection) can be accessed by the 'tab' hash table,
    which is indexed on each link's name field. Because there can be multiple
    objects with the same name, the hash table points to the
    first such object. Subsequent ones can be found by following
    the 'next' field (until next is NULL or next->name differs).
 */

struct corcollection {
    struct corSymbolTable tab[1];        // hash table
    Link *chain;        // list of rules (linked by each rule's next field)
    Link **lastptr;
    int count;
    int hashable;
};

/*****************************************************************************/

#endif _Collection_H_

