// *************************************************************************
//    SemiNaive.C
// *************************************************************************

#include "SemiNaive.h"

#include "SccInfo.h"
#include "PredAnnotations.h"
#include "CompileRule.h"
#include "Compile.h"
#include "Relation.h"
#include "Backtrack.h"
#include "Unparse.h"


// *************************************************************************
corSemiNaive::corSemiNaive()
{
    scc = NULL;
    rInfo = NULL;
    first = 0;
    markOffsets = NULL;
    negated = NULL;
    predOrders = NULL;
    firstFailureBacktrack = NULL;
    nextFailureBacktrack = NULL;  
    successBacktrack = 0;
    recursive = 0;
    indices = NULL;
    ignore_this_rule = 0;
    rule_id = -1;
    label = NULL;
}

// *************************************************************************
corSemiNaive::~corSemiNaive()
{
    if (rInfo) {
        if (markOffsets) delete markOffsets;
        if (indices) delete [] indices;
        if (negated) delete [] negated;
        if (predOrders) delete [] predOrders;
        if (rInfo->num_literals) {
            delete [] firstFailureBacktrack;
            delete [] nextFailureBacktrack;
        }
    }
    delete [] label;
}

// *************************************************************************
void corSemiNaive::clean_cache() 
{
    for (int i=0; i < rInfo->num_literals; i++)
        indices[i] = -1;
}

// *************************************************************************
void corSemiNaive::Add_Rule_Info(corRuleInfo *rule_info,
                  corClause *clause, int non_ground_facts)
{
    int i;

    rInfo = rule_info;
    markOffsets = new int[(rInfo->num_literals)*2];

    indices = new int [rInfo->num_literals];

    for (i = 0; i < rInfo->num_literals; i++)
        indices[i] = -1;

    predOrders = new int[rInfo->num_literals];
    corinitPredOrders(this);
        
    //    Generate index annotations if the SemiNaive rule is a part of an scc.
    //    In this case, the scc belongs to a module, and index annotations
    //    can be created for the module. However, if the rule is a single
    //    imperative rule (executed from the CORAL prompt), there is no
    //    module for it. A separate index annotation list is created and
    //    used to store the index annotations. These annotations are executed
    //    right away resulting in eager index creation.
    //    NOTE:: generateIndexes can change the PredOrders array ! Hence
    //    anything that uses the predOrders should come after it.      

    if (scc)
        corgenerateIndexes(this, scc->module->index_annotations);
    else {
        corPredAnnotations *tmp_anno = NULL, *cur;
        corgenerateIndexes(this, tmp_anno);

        for(cur= tmp_anno; cur;) {
            corVarLink *rest;
            rest = cur->arglist1->var_list(NULL);
            if ( cur->arglist2 )
                rest = cur->arglist2->var_list(rest);
            
            int num_vars = corcount_var_list(rest);
            corRelation *indexrel = corfind_external_relation(cur->pred,
                cur->arglist1->arity());
            if (indexrel) {
                if (cur->arglist2)
                    indexrel->add_index(cur->arglist1, num_vars, cur->arglist2);
                else
                    indexrel->add_index(cur->arglist1);
            }
            // delete space for index annos
            corPredAnnotations *temp = cur;
            cur = cur->next();
            delete temp;
        }    
        // Just deleted it above
        //delete tmp_anno;
    }

    //    NOTE: predOrders MUST be assigned BEFORE the backtrack arrays
    //    and negated arrays

    negated = new int[rInfo->num_literals];
    for (i = 0; i < rInfo->num_literals; i++)
        negated[i] = (rInfo->relations[i] < 0);

    if (rInfo->num_literals) {
        firstFailureBacktrack = new int [rInfo->num_literals];
        nextFailureBacktrack = new int [rInfo->num_literals];
        initFailureBacktrack(this, non_ground_facts);
        btDecision[BACKWARD] = nextFailureBacktrack;
        btDecision[FORWARD] = firstFailureBacktrack;
        successBacktrack =
            corgetSuccessBacktrackPoint(this,clause,non_ground_facts);
    }
    
    clean_cache();

    // this is an optimization for supplementary magic with indexing
    // the magic rules are ignored, and the magic facts are generated
    // along withthe firing of the goalid() literal of the supp rule
    //              --- Praveen
    // note that scc could be NULL if this is an imperative rule invoked
    // from the coral prompt
    /******* This causes problems with return unify optimization : Praveen
    if (scc && (scc->module->module_info.SupMagicIndexing) && 
          is_magic(clause->head->pred))
        ignore_this_rule = 1;
    *****/
}

// *************************************************************************
void corSemiNaive::print(FILE *fd)
{ // print rule info
  //  rInfo->pInfo.print(fd, "", predOrders, rInfo->num_literals);
  fprintf(fd, "  sn_rule #%d: ", rule_id);
  fprintf(fd, "%s: ", label);
  corprint_rule_with_order(rInfo->rule, predOrders, markOffsets, recursive, fd);
    
}

// *************************************************************************
void corSemiNaive::Add_label()
{
  label = new char[20];

  int org_rule_no, magic_rule_no;
  
  // get the original rule number
  if (sscanf(rInfo->rule->clause_names->name->string(), "%d", 
           &org_rule_no) < 1)
      org_rule_no = 0;

  // get the magic rule number, if it has one
  if (rInfo->rule->clause_names->next){
     if ((sscanf(rInfo->rule->clause_names->next->name->string(), "%d", 
           &magic_rule_no)) < 1)
        magic_rule_no = 0;
  } else
    magic_rule_no = 0;

  // form a label for the SN rule using the original rule no, magic rule no
  // and the label counter
  sprintf(label, "%d.%d.%d", org_rule_no, magic_rule_no, 
      rInfo->rule->sn_rule_label_counter);
  rInfo->rule->sn_rule_label_counter++;

}

// *************************************************************************
