/*****************************************************************************/
//	Defunct.C
/*****************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "Defunct.h"

#include "TermList.h"
#include "Print.h"
#include "FunctArgs.h"

#include "Globals.h"
#include "Rule.h"
#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "DontCareArg.h"
#include "CArg.h"
/*****************************************************************************/

extern FILE *outf;
extern int orig_rule_count;

/*****************************************************************************/

static int var_count = 0;
static int duplicate_var_count = 0;

#define TRACE(x) if( 0 ) { fprintf x ; fflush( stderr ) ;};

/*****************************************************************************/

static int add_variable(corClause *rule, Name name);
static void add_equ_literal(corClause *rule, const corCArg&, const corCArg &);
static void replace_equ_literal(corClause *rule, corLiteral *lit);
static corArg* make_new_var(corClause *rule, const corCArg&);
static void make_distinct_vars(corClause *rule, corLiteral *lit);
static void make_distinct_vars(corClause *rule);
static void collect_terms(corClause *rule , TermList *terms);
static void explode(TermList *terms);
static void associate_vars(corClause * rule, TermList *terms);
static void freeup(TermList *terms);
static void reverse(TermList *terms);
static void reverse_list(TermList *terms);
static corFuncArg * in_termlist(TermList *terms, corVarArg *a, corLiteral * &pred);
static corFuncArg * in_termlist(TermList **terms, corVarArg *a, corLiteral * &pred);
static void remove_special_pred(corClause *rule, corLiteral *pred);
static void restore_arglist(corClause *rule,TermList **terms, corArgList &alist,
	corLiteral *p);

/*****************************************************************************/
void defunctify(corClause *rules, TermList **terms)
{
	corClause *r;
	int i;

	//	Process rules
	for (i=0,r=rules; r; r=(corClause *)(r->link.next), i++) {
		// This check is just to ensure that the TermList array
		// has one slot for every rule.
		ASSERT( i < orig_rule_count);

		/*
			Ensure that all non-functor arguments are variables
			by introducing additional equ_(X,Y) literals on the rhs.
			NOTE:: this should actually be handled along with
			functors in an identical fashion. This current approach is
			only a kludge --- PRAVEEN
		*/
		make_distinct_vars(r);

		// collect set of all terms with function symbols
		terms[i] = new TermList;
		collect_terms(r, terms[i]);

		//	Generate subterms for the set of subterms and associate
		//	using forward pointers.
		explode(terms[i]);

		//	Reverse the term list, so that nested funcargs are expanded
		//	in the right order.
		reverse_list(terms[i]);

		//	Do a scan from smallest subterm, and associate with each term
		//	a unique variable symbol.
		associate_vars(r, terms[i]);
	}
}

/*****************************************************************************/
static int add_variable(corClause *rule, Name name)
{
	rule->var_names = (Name *)realloc(rule->var_names,
		++(rule->num_var_names) * sizeof(Name));
	rule->var_names[rule->num_var_names - 1] = name;
	return rule->num_var_names - 1;
}

/*****************************************************************************/
/*
	Add a literal of the form equ_(old_arg, new_vararg).
*/

static void add_equ_literal(corClause *rule, const corCArg& arg0, const corCArg& arg1)
{
	corLiteral *tmp2 = new corLiteral(corparserStruct, corEquOpSymbol, 2);
						 
	tmp2->adorn.clear();
	tmp2->args[0] = arg0;
	tmp2->args[1] = arg1;
	tmp2->base_pred = 1;
	rule->preds = (corLiteral **)realloc(rule->preds,
		++(rule->num_preds) * sizeof(corLiteral *));
	rule->preds[rule->num_preds - 1] = tmp2;
}

/*****************************************************************************/
static void replace_equ_literal(corClause *rule, corLiteral *lit)
{
	int j;

	ASSERT(lit != NULL);
	if (lit->name() != corEquOpSymbol) {
		//	This should not happen
		return;
	}

	for (j = 0; j < rule->head->args.arity(); j++) {
		if (rule->head->args[j] == lit->args[1]) {
			rule->head->args[j] = lit->args[0];
			return;
		}
	}
	
	for (int i = 0; i < rule->num_preds; i++) {
		if (rule->preds[i] == lit) break;
		for (j = 0; j < rule->preds[i]->args.arity(); j++)
			if (rule->preds[i]->args[j] == lit->args[1]) {
				rule->preds[i]->args[j] = lit->args[0];
				return;
			}
	}
}

/*****************************************************************************/
static corArg* make_new_var(corClause *rule, const corCArg& c)
{
	char buf[100];
	Name name = c.SymbolP();
	sprintf(buf, "V%s__%d", corSymbolString(name), var_count++);
	Name tmp1 = corEnterSymbol(buf);
	return (new corVarArg(tmp1, add_variable(rule, tmp1)));
}

/*****************************************************************************/
static void make_distinct_vars(corClause *rule, corLiteral *lit)
{
	ASSERT( lit != NULL );

	int j = 0;
	corCArg a;
	for (j = 0; j < (lit->args).arity(); j++) {
		a = (lit->args)[j];
		switch (((lit->args)[j]).kindof()) {
			case COR_NUM_CONST:
			case COR_SYMBOL_CONST:
			case COR_RELATION:
			case COR_GROUPING:
				a = (lit->args)[j];
				(lit->args)[j] = make_new_var(rule, lit->name());
				duplicate_var_count++;
				add_equ_literal(rule, a, (lit->args)[j]);
				break;
			case COR_VARIABLE:
			case COR_DONTCARE: {
				int i;
				for (i = j+1; i < (lit->args).arity(); i++) {
					if ((lit->args)[i].VarArgP()->var_name() ==
						((lit->args)[j].VarArgP())->var_name())
					{
						a = lit->args[i];
						lit->args[i] = make_new_var(rule, lit->name());
						duplicate_var_count++;
						add_equ_literal(rule, a, (lit->args)[i]);
					}
					else 
						break;
				}
			}
			default:
				break;
		}
	}
}

/*****************************************************************************/
static void make_distinct_vars(corClause *rule)
{
	int i;

	duplicate_var_count = 0;
	make_distinct_vars(rule, rule->head);
	for (i = 0; i < rule->num_preds; i++) {
		if (rule->preds[i]->name() != corEquOpSymbol)
			make_distinct_vars(rule, rule->preds[i]);
	}

	//	Ghastly kludge to store the number of extra predicates created in
	//	this phase.  old_num_var_names is a field that is not being used,
	//	so i'm using it for this purpose -- PRAVEEN
	rule->old_num_var_names = duplicate_var_count;
}

/*****************************************************************************/
/*
This only collects top level terms that are arguments of predicates in a rule 
*/

static void collect_terms(corClause *rule , TermList *terms)
{
	int i;

	get_terms(rule->head, terms );
	TRACE((stderr, "Got terms for head\n"));
	for(i = 0; i < rule->num_preds; i++) {
		get_terms( rule->preds[i], terms );
		TRACE((stderr, "Got new set of terms for body pred\n" ) );
	}
	TRACE((stderr, "finished collecting terms\n"));
}

/*****************************************************************************/
/*
	When exploding a term, we look at each sub term.  If the subterm is 
	already present in the final list, we add a forward pointer from
	the term to the unique copy in the final list.  If it is not present,
	we copy the subterm and introduce it at the end of final list.  We also 
	introduce a forward pointer from the subterm to the copy in the
	final list.  We iteratively end up exploding all subterms since they
	are put at the end of final list and are eventually processed by this
	procedure.
*/

static void explode(TermList *terms)
{
	TermList *tmp_terms = terms->next;

	for (; tmp_terms; tmp_terms = tmp_terms->next) {
		ASSERT(tmp_terms->func_arg->kindof() == COR_FUNCTOR);
		int j = 0;
		FOR_EACH_ARG(a, tmp_terms->func_arg->args) {
			if (a->kindof() == COR_FUNCTOR) {
				TermList *tmp = find_arg(terms, a->FuncArgP());
				add_arg(terms,a->FuncArgP(), &(((corFuncArg *)(tmp_terms->func_arg))->args), j, tmp);
			}
			j++;
		} END_EACH_ARG
	}
	TRACE((stderr, "Exploded term list\n"));
}

/*****************************************************************************/
/*
	Associates a new variable with each occurrence of a term in the 
	exploded term list associated with the rule.  
*/

static void associate_vars(corClause * rule, TermList *terms)
{
	char buf[100];

	for (terms = terms->next; terms; terms = terms->next) {
		if (terms->similar) {
			TermList *tmp = terms->similar;
			while (tmp->similar) tmp = tmp->similar;
			(*((corArgList *)(terms->arg_list)))[terms->arg_pos] = tmp->var_arg;
			terms->var_arg = tmp->var_arg;
			continue;
		}			

		sprintf(buf, "%s__%d", corSymbolString(terms->func_arg->functor()),
			var_count);

		Name tmp1 = corEnterSymbol(buf);
		terms->var_arg = new corVarArg(tmp1, add_variable(rule, tmp1));
		(*((corArgList *)(terms->arg_list)))[terms->arg_pos] = terms->var_arg;

		//	Now add an extra predicate to the rule
		//	COR_NOTE :: Forced to use realloc() here instead of new() because
		//	alloc_preds() initially used malloc !!! --- PRAVEEN

		sprintf(buf, "func__%d", var_count++); 
		corLiteral *tmp2 = new corLiteral(corparserStruct, corEnterSymbol(buf),
			terms->func_arg->count()+1);
		tmp2->adorn.clear();
		for (int k=0; k< terms->func_arg->count();k++)
			tmp2->args[k] = terms->func_arg->args[k];
		tmp2->args[k] = terms->var_arg;
		tmp2->base_pred = 1;
		rule->preds = (corLiteral **)realloc(rule->preds,
			++(rule->num_preds) * sizeof(corLiteral *));
		rule->preds[rule->num_preds - 1] = tmp2;
		terms->dummy_pred = tmp2;
	}

	TRACE((stderr, "Finished associating vars\n"));
}

/*****************************************************************************/
static void freeup(TermList *terms)
{
	if (!terms) return;
	while (terms) {
		TermList *tmp = terms->next;
		delete terms;
		terms = tmp;
	}
}

/*****************************************************************************/
static void reverse(TermList *terms)
{
	if (terms->next == NULL) return;
	reverse(terms->next);
	terms->next->next = terms;
	terms->next = NULL;
}

/*****************************************************************************/
static void reverse_list(TermList *terms)
{
	TermList *tmp = terms->next;
	if (!tmp) return;
	while (tmp->next) tmp = tmp->next;
	reverse(terms->next);
	terms->next = tmp;
}

/*****************************************************************************/
static corFuncArg * in_termlist(TermList *terms, corVarArg *a, corLiteral * &pred)
{
	if (!terms)	return NULL;

	for (terms = terms->next; terms; terms= terms->next) {
		if (a->var_name() == terms->var_arg->var_name()) {
			pred = terms->dummy_pred;
			return terms->func_arg;
		}
	}
	return NULL;
}

/*****************************************************************************/
static corFuncArg * in_termlist(TermList **terms, corVarArg *a, corLiteral * &pred)
{
	for (int j=0; j< orig_rule_count; j++) {
		 corFuncArg * tmp_funcarg = in_termlist(terms[j], a, pred);
		 if (tmp_funcarg)
			return tmp_funcarg;
	}
	return NULL;
}

/*****************************************************************************/
static void remove_special_pred(corClause *rule, corLiteral *pred)
{
	for (int i = 0; i< rule->num_preds; i++) {
	if ((pred->name() == rule->preds[i]->name()) &&
		pred->arity() == rule->preds[i]->arity()) break;
	}

	// If not found, return ---- this should not happen !!
	if (i == rule->num_preds) return;

	// Copy the adornment in the literal 
	pred->adorn = rule->preds[i]->adorn;

	// Remove the literal from the rule
	for (; i< rule->num_preds; i++) rule->preds[i] = rule->preds[i++];
	(rule->num_preds)--;
}

/*****************************************************************************/
//	Replace dummy variables by original func args.

static void restore_arglist(corClause *rule,TermList **terms, corArgList &alist,
	corLiteral *p)
{
	int k = 0;
	corFuncArg * f_arg;
	FOR_EACH_ARG(a, alist) {
		if (p && p->adorn[k])
			alist[k] = &corTheDontCareArg;

		if (a->kindof() != COR_VARIABLE) continue;

		//	At this point, all args must be of type VARARG
		ASSERT(a->kindof() == COR_VARIABLE);
		corLiteral *tmp_pred = NULL;
		if (f_arg = in_termlist(terms, a->VarArgP(), tmp_pred)) {
			remove_special_pred(rule, tmp_pred);
			if (alist[k] != &corTheDontCareArg) {
				alist[k] = f_arg;
				restore_arglist(rule, terms, f_arg->args, tmp_pred);
			}
		}
		k++;
	} END_EACH_ARG
}

/*****************************************************************************/
void functify(corClause *rules, TermList **terms)
{
	int i, j;
	corClause * r;

	//	Process rules 

	for (i=0,r=rules; r; r=(corClause *)(r->link.next), i++) {
		//	Process predicates
		restore_arglist(r, terms, r->head->args, NULL);

		for (j=0; j< r->num_preds; j++)
			restore_arglist(r, terms, r->preds[j]->args, NULL);

		for (j=0; j< r->old_num_var_names; j++)
			replace_equ_literal(r, r->preds[r->num_preds - j -1]);

		r->num_preds -= r->old_num_var_names;
	}
}

/*****************************************************************************/

